#!/usr/bin/env jupyter

import typing as t
from abc import ABC
from os.path import join

import numpy as np
from skimage.measure import regionprops_table

from aliby.track.utils import calc_barycentre, pick_baryfun


class FeatureCalculator(ABC):
    """
    Base class for making use of regionprops-based features.
    If no features are offered it uses most of them.

    This class is not to be used directly
    """

    a_ind = None
    ma_ind = None
    x_ind = None
    y_ind = None

    def __init__(
        self,
        feats2use: t.Collection[str],
        trapfeats: t.Optional[t.Collection[str]] = None,
        extrafeats: t.Optional[t.Collection[str]] = None,
        aweights: t.Optional[bool] = None,
        pixel_size: t.Optional[float] = None,
    ) -> None:

        self.feats2use = feats2use

        if trapfeats is None:
            trapfeats = ()
        self.trapfeats = tuple(trapfeats)

        if extrafeats is None:
            extrafeats = ()
        self.extrafeats = tuple(extrafeats)

        if aweights is None:
            aweights = None
        self.aweights = aweights

        if pixel_size is None:
            pixel_size = 0.182
        self.pixel_size = pixel_size

        self.outfeats = self.get_outfeats()

        self.set_named_ids()

        self.tfeats = self.outfeats + self.tmp_outfeats + self.trapfeats
        self.ntfeats = len(self.tfeats)

    def get_outfeats(
        self, feats2use: t.Optional[t.Collection[str]] = None
    ) -> tuple:
        if feats2use is None:
            feats2use = self.feats2use
        outfeats = tuple(
            regionprops_table(np.diag((1, 0)), properties=feats2use).keys()
        )
        return outfeats

    def set_named_ids(self):
        # Manage calling feature outputs by name
        d = {"centroid-0": "xind", "centroid-1": "yind", "area": "aind"}
        tmp_d = {
            "barydist": ["centroid", "area"],
            "baryangle": ["centroid", "area"],
            "distance": ["centroid"],
        }

        nonbase_feats = self.trapfeats + self.extrafeats

        tmp_infeats = np.unique([j for x in nonbase_feats for j in tmp_d[x]])
        self.tmp_infeats = tuple(
            [f for f in tmp_infeats if f not in self.feats2use]
        )

        # feats that are only used to calculate others
        tmp_outfeats = (
            self.get_outfeats(feats2use=tmp_infeats)
            if len(tmp_infeats)
            else []
        )

        self.tmp_outfeats = []
        for feat in tmp_outfeats:
            if feat in self.outfeats:
                setattr(self, d[feat], self.outfeats.index(feat))
            else:  # Only add them if not in normal outfeats
                self.tmp_outfeats.append(feat)
                setattr(
                    self,
                    d[feat],
                    len(self.outfeats) + self.tmp_outfeats.index(feat),
                )

        self.tmp_outfeats = tuple(self.tmp_outfeats)
        self.out_merged = self.outfeats + self.tmp_outfeats

    def load_model(self, path, fname):
        model_file = join(path, fname)
        with open(model_file, "rb") as file_to_load:
            model = pickle.load(file_to_load)

        return model

    def calc_feats_from_mask(
        self,
        masks: np.ndarray,
        feats2use: t.Optional[t.Tuple[str]] = None,
        trapfeats: t.Optional[t.Tuple[str]] = None,
        scale: t.Optional[bool] = True,
        pixel_size: t.Optional[float] = None,
    ):
        """
        Calculate feature ndarray from ndarray of cell masks
        ---
        input

        :masks: ndarray (ncells, x_size, y_size), typically dtype bool
        :feats2use: list of strings with the feature properties to extract.
            If it is None it uses the ones set in self.feats2use.
        :trapfeats: List of str with additional features to use
            calculated immediately after basic features.
        :scale: bool, if True scales mask to a defined pixel_size.
        :pixel_size: float, used to rescale the object features.

        returns

        (ncells, nfeats) ndarray of features for input masks
        """
        if pixel_size is None:
            pixel_size = self.pixel_size

        if feats2use is None:
            feats2use = self.feats2use + self.tmp_infeats

        if trapfeats is None:
            trapfeats = self.trapfeats

        ncells = masks.shape[0] if masks.ndim == 3 else masks.max()
        feats = np.empty((ncells, self.ntfeats))  # ncells * nfeats
        if masks.any():
            if masks.ndim == 3:  # Individual cells in dim 0
                assert masks.sum(
                    axis=(1, 2)
                ).all(), "Dimension with at least one empty outline slice"

                cell_feats = np.array(
                    [
                        [
                            x[0]
                            for x in regionprops_table(
                                mask.astype(int), properties=feats2use
                            ).values()
                        ]
                        for mask in masks
                    ]
                )

            elif masks.ndim == 2:  # No overlap between cells
                cell_feats = np.array(
                    [
                        x
                        for x in regionprops_table(
                            masks.astype(int), properties=feats2use
                        ).values()
                    ]
                ).T
            else:
                raise Exception(
                    "TrackerException: masks do not have the appropiate dimensions"
                )

            if scale:
                cell_feats = self.scale_feats(cell_feats, pixel_size)

            # Fill first sector, with directly extracted features
            feats[:, : len(self.out_merged)] = cell_feats
            if trapfeats:  # Add additional features

                tfeats = self.calc_trapfeats(feats)
                feats[:, len(self.out_merged) :] = tfeats
        else:

            feats = np.zeros((0, self.ntfeats))

        return feats

    def calc_trapfeats(self, basefeats):
        """
        Calculate trap-based features
        using basic ones.
        :basefeats: (n basic outfeats) 1-D array with features outputed by
            skimage.measure.regionprops_table


        requires
            self.aind
            self.aweights
            self.xind
            self.yind
            self.trapfeats

        returns
        (ntrapfeats) 1-D array with
        """
        if self.aweights is not None:
            weights = basefeats[:, self.aind]
        else:
            weights = None

        barycentre = calc_barycentre(
            basefeats[:, [self.xind, self.yind]], weights=weights
        )

        trapfeat_nd = np.empty((basefeats.shape[0], len(self.trapfeats)))
        for i, trapfeat in enumerate(self.trapfeats):
            trapfeat_nd[:, i] = pick_baryfun(trapfeat)(
                basefeats[:, [self.xind, self.yind]], barycentre
            )

        return trapfeat_nd

    def scale_feats(self, feats: np.ndarray, pixel_size: float):
        """
        input

        :feats: np.ndarray (ncells * nfeatures)
        :pixel_size: float Value used to normalise the images.

        returns
        Rescaled list of feature values
        """
        area = pixel_size**2

        scaling = {None: 1, "linear": pixel_size, "square": area}
        degrees_feats = {
            None: ["eccentricity", "extent", "orientation", "solidity"],
            "linear": [
                "centroid-0",
                "centroid-1",
                "minor_axis_length",
                "major_axis_length",
                "perimeter",
                "perimeter_crofton",
                "equivalent_diameter",
            ],
            "square": [
                "area",
                "convex_area",
                "bbox_area",
                "equivalent_diameter",
            ],
        }

        scaler = {
            feat: scaling[k]
            for k, feats in degrees_feats.items()
            for feat in feats
        }
        # for k in feats.keys():
        #     feats[k] /= scaler[k]

        return feats * [scaler[feat] for feat in self.out_merged]



#!/usr/bin/env jupyter
"""
Add general logging functions and decorators
"""

import logging
from time import perf_counter


def timer(func):
    # Log duration of a function into aliby logfile
    def wrap_func(*args, **kwargs):
        t1 = perf_counter()
        result = func(*args, **kwargs)
        logging.getLogger("aliby").debug(
            f"{func.__qualname__} took {(perf_counter()-t1):.4f}s"
        )
        return result

    return wrap_func



"""
Functions to load and reshape examples for extraction development

The basic format for data will be pair data/masks pairs. Data will be
assumed to be a single slide, given that this reduction is expected
to happen beforehand.

The most basic functions were copied from Swain Lab's baby module,
specifically baby/io.py
"""

from importlib_resources import files
import json
import re
from itertools import groupby
from pathlib import Path
from typing import Callable

import numpy as np
from imageio import imread

from extraction.core.functions.distributors import reduce_z


def load_tiled_image(filename):
    tImg = imread(filename)
    info = json.loads(tImg.meta.get("Description", "{}"))
    tw, th = info.get("tilesize", tImg.shape[0:2])
    nt = info.get("ntiles", 1)
    nr, nc = info.get("layout", (1, 1))
    nc_final_row = np.mod(nt, nc)
    img = np.zeros((nt, tw, th), dtype=tImg.dtype)
    for i in range(nr):
        i_nc = nc_final_row if i + 1 == nr and nc_final_row > 0 else nc
        for j in range(i_nc):
            ind = i * nc + j
            img[ind, :, :] = tImg[i * tw : (i + 1) * tw, j * th : (j + 1) * th]

    return img, info


def load_paired_images(filenames, typeA="Brightfield", typeB="segoutlines"):
    re_imlbl = re.compile(r"^(.*)_(" + typeA + r"|" + typeB + r")\.png$")
    # For groupby to work, the list needs to be sorted; also has the side
    # effect of ensuring filenames is no longer a generator
    filenames = sorted(filenames)
    matches = [re_imlbl.match(f.name) for f in filenames]
    valid = filter(lambda m: m[0], zip(matches, filenames))
    grouped = {
        k: {m.group(2): f for m, f in v}
        for k, v in groupby(valid, key=lambda m: m[0].group(1))
    }
    valid = [
        set(v.keys()).issuperset({typeA, typeB}) for v in grouped.values()
    ]
    if not all(valid):
        raise Exception
    return {
        lbl: {t: load_tiled_image(f) for t, f in g.items()}
        for lbl, g in grouped.items()
    }


def load(path=None):
    """
    Loads annotated pngs into memory. Only designed for GFP and brightfield.

    input
    :path: Folder used to look for images

    returns
    list of dictionaries containing GFP, Brightfield and segoutlines channel
    """
    if path is None:

        path = (
            files("aliby").parent.parent
            / "examples"
            / "extraction"
            / "pairs_data"
        )

    image_dir = Path(path)
    channels = ["Brightfield", "GFP"]
    imset = {"segoutlines": {}}
    for ch in channels:
        imset[ch] = {}
        pos = load_paired_images(image_dir.glob("*.png"), typeA=ch)
        for key, img in pos.items():
            imset[ch][key] = img[ch][0]
            imset["segoutlines"][key] = img["segoutlines"][0].astype(bool)

    return [{ch: imset[ch][pos] for ch in imset.keys()} for pos in pos.keys()]


def load_1z(fun: Callable = np.maximum, path: str = None):
    """
    ---
        fun: Function used to reduce the multiple stacks
        path: Path to pass to load function


    """
    dsets = load(path=path)
    dsets_1z = []
    for dset in dsets:
        tmp = {}
        for ch, img in dset.items():
            if ch == "segoutlines":
                tmp[ch] = img
            else:
                tmp[ch] = reduce_z(img, fun)

        dsets_1z.append(tmp)

    return dsets_1z



""" How to do the nuc Est Conv from MATLAB
Based on the code in MattSegCode/Matt Seg
GUI/@timelapseTraps/extractCellDataStacksParfor.m

Especially lines 342 to  399.
This part only replicates the method to get the nuc_est_conv values
"""
import typing as t
import numpy as np
import skimage
from scipy import signal, stats


def matlab_style_gauss2D(shape=(3, 3), sigma=0.5):
    """
    2D gaussian mask - should give the same result as MATLAB's
    fspecial('gaussian',[shape],[sigma])
    """
    m, n = [(ss - 1.0) / 2.0 for ss in shape]
    y, x = np.ogrid[-m : m + 1, -n : n + 1]
    h = np.exp(-(x * x + y * y) / (2.0 * sigma * sigma))
    h[h < np.finfo(h.dtype).eps * h.max()] = 0
    sumh = h.sum()
    if sumh != 0:
        h /= sumh
    return h


def gauss3D(
    shape: t.Tuple[int] = (3, 3, 3), sigma: t.Tuple[float] = (0.5, 0.5, 0.5)
):
    """3D gaussian mask - based on MATLAB's fspecial but made 3D."""
    m, n, p = [(ss - 1.0) / 2.0 for ss in shape]
    z, y, x = np.ogrid[-p : p + 1, -m : m + 1, -n : n + 1]
    sigmax, sigmay, sigmaz = sigma
    xx = (x**2) / (2 * sigmax)
    yy = (y**2) / (2 * sigmay)
    zz = (z**2) / (2 * sigmaz)
    h = np.exp(-(xx + yy + zz))
    h[h < np.finfo(h.dtype).eps * h.max()] = 0  # Truncate
    sumh = h.sum()
    if sumh != 0:
        h /= sumh
    return h


def small_peaks_conv(cell_mask: np.ndarray, trap_image: np.ndarray):
    cell_fluo = trap_image[cell_mask]
    # Get the number of pixels in the cell
    num_cell_fluo = len(np.nonzero(cell_fluo)[0])
    # Sort cell pixels in descending fluorescence order
    ratio_overlap = num_cell_fluo * 0.025  # TODO what is this?

    # Small Peak Conv
    # Convolution parameters
    conv_matrix = np.zeros((3, 3))
    # This makes a matrix with zeros in the corners and ones every where else
    # Basically the minimal disk.
    conv_matrix[1, :] = 1
    conv_matrix[:, 1] = 1
    # Reshape so that it is the size of a fifth of the cell, which is what we
    # expect the size of the nucleus to be.
    # TODO directly get a disk of that size?
    # new_shape = tuple(x * ratio_overlap / 5 for x in conv_matrix.shape)
    # conv_matrix = misc.imresize(conv_matrix, new_shape)
    conv_matrix = skimage.morphology.disk(3 * ratio_overlap / 5)
    # Apply convolution to the image
    # TODO maybe rename 'conv_matrix' to 'kernel'
    fluo_peaks = signal.convolve(trap_image, conv_matrix, "same")
    fluo_peaks = fluo_peaks[cell_mask]
    small_peak_conv = np.max(fluo_peaks)
    return small_peak_conv


def nuc_est_conv(
    cell_mask: np.ndarray,
    trap_image: np.ndarray,
    alpha: t.Optional[float] = 0.95,
    object_radius_estimation: t.Optional[float] = 0.085,
    gaussian_filter_shape: t.Optional[t.Union[int, t.Tuple[int]]] = None,
    gaussian_sigma: t.Optional[float] = None,
):
    """
    :param cell_mask: the segmentation mask of the cell (filled)
    :param trap_image: the image for the trap in which the cell is (all
    channels)
    :param alpha: optional distribution alpha to get confidence intervals
    :param object_radius_estimation: optional estimated object volume
    (in pixels), used to estimate the object radius.
    :param gaussian_filter_shape: optional tuple to pass to matlab_style_gauss2D,
    determines the kernel shape for convolutions.
    :param gaussian_sigma: optional optional sigma to pass to matlab_style_gauss2D
    as sigma argument.
    """
    if alpha is None:
        alpha = 0.95

    if object_radius_estimation is None:
        object_radius_estimation = 0.085

    cell_loc = cell_mask  # np.where(cell_mask)[0]
    cell_fluo = trap_image[cell_mask]
    num_cell_fluo = len(np.nonzero(cell_fluo)[0])

    chi2inv = stats.distributions.chi2.ppf(alpha, df=2)

    approx_nuc_radius = np.sqrt(
        object_radius_estimation * num_cell_fluo / np.pi
    )

    if gaussian_sigma is None:
        gaussian_sigma = float(approx_nuc_radius / np.sqrt(chi2inv))

    # Nuc Est Conv
    filter_size = int(np.ceil(2 * approx_nuc_radius))
    gaussian_filter_shape = (2 * filter_size + 1,) * 2

    nuc_filter = matlab_style_gauss2D(gaussian_filter_shape, gaussian_sigma)

    cell_image = trap_image - np.median(cell_fluo)
    cell_image[~cell_loc] = 0

    nuc_conv = signal.convolve(cell_image, nuc_filter, "same")
    nuc_est_conv = np.max(nuc_conv)
    nuc_est_conv /= (
        np.sum(nuc_filter**2) * alpha * np.pi * chi2inv * gaussian_sigma**2
    )
    return nuc_est_conv


def nuc_conv_3d(cell_mask, trap_image, pixel_size=0.23, spacing=0.6):
    cell_mask = np.stack([cell_mask] * trap_image.shape[0])
    ratio = spacing / pixel_size
    cell_fluo = trap_image[cell_mask]
    num_cell_fluo = len(np.nonzero(cell_fluo)[0])
    # Nuc Est Conv
    alpha = 0.95
    approx_nuc_radius = np.sqrt(0.085 * num_cell_fluo / np.pi)
    chi2inv = stats.distributions.chi2.ppf(alpha, df=2)
    sd_est = approx_nuc_radius / np.sqrt(chi2inv)
    nuc_filt_hw = np.ceil(2 * approx_nuc_radius)
    nuc_filter = gauss3D(
        (2 * nuc_filt_hw + 1,) * 3, (sd_est, sd_est, sd_est * ratio)
    )
    cell_image = trap_image - np.median(cell_fluo)
    cell_image[~cell_mask] = 0
    nuc_conv = signal.convolve(cell_image, nuc_filter, "same")
    nuc_est_conv = np.max(nuc_conv)
    nuc_est_conv /= (
        np.sum(nuc_filter**2) * alpha * np.pi * chi2inv * sd_est**2
    )
    return nuc_est_conv



from agora.abc import ParametersABC

from postprocessor.core.abc import PostProcessABC
from postprocessor.core.functions.tracks import get_joinable


class MergerParameters(ParametersABC):
    """
    Define the parameters for merger from a dict.

    There are five parameters expected in the dict:

    smooth, boolean
        Whether or not to smooth with a savgol_filter.
    tol: float or  int
        The threshold of average prediction error/std necessary to
        consider two tracks the same.
        If float, the threshold is the fraction of the first track;
        if int, the threshold is in absolute units.
    window: int
        The size of the window of the savgol_filter.
    degree: int v
        The order of the polynomial used by the savgol_filter
    """

    _defaults = {
        "smooth": False,
        "tolerance": 0.2,
        "window": 5,
        "degree": 3,
        "min_avg_delta": 0.5,
    }


class Merger(PostProcessABC):
    """Combine rows of tracklet that are likely to be the same."""

    def __init__(self, parameters):
        super().__init__(parameters)

    def run(self, signal):
        joinable = []
        if signal.shape[1] > 4:
            joinable = get_joinable(
                signal,
                smooth=self.parameters.smooth,
                tol=self.parameters.tolerance,
                window=self.parameters.window,
                degree=self.parameters.degree,
            )
        return joinable



import typing as t

import numpy as np
import pandas as pd

from agora.abc import ParametersABC
from agora.io.cells import Cells
from agora.utils.indexing import validate_association
from agora.utils.cast import _str_to_int
from agora.utils.kymograph import drop_mother_label
from postprocessor.core.lineageprocess import LineageProcess


class PickerParameters(ParametersABC):
    _defaults = {
        "sequence": [
            ["lineage", "families"],
            ["condition", "present", 7],
        ],
    }


class Picker(LineageProcess):
    """
    :cells: Cell object passed to the constructor
    :condition: Tuple with condition and associated parameter(s), conditions can be
    "present", "nonstoply_present" or "quantile".
    Determine the thresholds or fractions of signals to use.
    :lineage: str {"mothers", "daughters", "families" (mothers AND daughters), "orphans"}. Mothers/daughters picks cells with those tags, families pick the union of both and orphans the difference between the total and families.
    """

    def __init__(
        self,
        parameters: PickerParameters,
        cells: Cells or None = None,
    ):
        super().__init__(parameters=parameters)
        self.cells = cells

    def pick_by_lineage(
        self,
        signal: pd.DataFrame,
        how: str,
        mothers_daughters: t.Optional[np.ndarray] = None,
    ) -> pd.MultiIndex:
        cells_present = drop_mother_label(signal.index)
        mothers_daughters = self.get_lineage_information(signal)
        valid_indices = slice(None)
        if how == "mothers":
            _, valid_indices = validate_association(
                mothers_daughters, cells_present, match_column=0
            )
        elif how == "daughters":
            _, valid_indices = validate_association(
                mothers_daughters, cells_present, match_column=1
            )
        elif how == "families":  # Mothers and daughters that are still present
            _, valid_indices = validate_association(
                mothers_daughters, cells_present
            )
        return signal.index[valid_indices]

    def pick_by_condition(self, signal, condition, thresh):
        idx = self.switch_case(signal, condition, thresh)
        return idx

    def run(self, signal):
        self.orig_signal = signal
        indices = set(signal.index)
        lineage = self.get_lineage_information(signal)
        if len(lineage):
            self.mothers = lineage[:, :2]
            self.daughters = lineage[:, [0, 2]]
            for alg, *params in self.sequence:
                new_indices = tuple()
                if indices:
                    if alg == "lineage":
                        param1 = params[0]
                        new_indices = getattr(self, "pick_by_" + alg)(
                            signal.loc[list(indices)], param1
                        )
                    else:
                        param1, *param2 = params
                        new_indices = getattr(self, "pick_by_" + alg)(
                            signal.loc[list(indices)], param1, param2
                        )
                        new_indices = [tuple(x) for x in new_indices]
                indices = indices.intersection(new_indices)
        else:
            self._log(f"No lineage assignment")
            indices = np.array([])
        return np.array([tuple(map(_str_to_int, x)) for x in indices])

    def switch_case(
        self,
        signal: pd.DataFrame,
        condition: str,
        threshold: t.Union[float, int, list],
    ):
        if len(threshold) == 1:
            threshold = [_as_int(*threshold, signal.shape[1])]
        case_mgr = {
            "any_present": lambda s, thresh: any_present(s, thresh),
            "present": lambda s, thresh: s.notna().sum(axis=1) > thresh,
            "nonstoply_present": lambda s, thresh: s.apply(thresh, axis=1)
            > thresh,
            "growing": lambda s, thresh: s.diff(axis=1).sum(axis=1) > thresh,
        }
        return set(signal.index[case_mgr[condition](signal, *threshold)])


def _as_int(threshold: t.Union[float, int], ntps: int):
    if type(threshold) is float:
        threshold = ntps * threshold
    return threshold


def any_present(signal, threshold):
    """
    Return a mask for cells, True if there is a cell in that trap that was present for more than :threshold: timepoints.
    """
    any_present = pd.Series(
        np.sum(
            [
                np.isin([x[0] for x in signal.index], i) & v
                for i, v in (signal.notna().sum(axis=1) > threshold)
                .groupby("trap")
                .any()
                .items()
            ],
            axis=0,
        ).astype(bool),
        index=signal.index,
    )
    return any_present



"""
Underlying methods for different neural network deployments.
"""
import itertools
import logging
import re
import time
import typing as t
from pathlib import Path
from time import perf_counter

import baby.errors
import h5py
import numpy as np
import requests
from agora.abc import ParametersABC, StepABC
from baby import modelsets
from baby.brain import BabyBrain
from baby.crawler import BabyCrawler
from requests.exceptions import HTTPError, Timeout


################### Dask Methods ################################
def format_segmentation(segmentation, tp):
    """Format a single timepoint into a dictionary.

    Parameters
    ------------
    segmentation: list
                  A list of results, each result is the output of the crawler, which is JSON-encoded
    tp: int
        the time point considered

    Returns
    --------
    A dictionary containing the formatted results of BABY
    """
    # Segmentation is a list of dictionaries, ordered by trap
    # Add trap information
    # mother_assign = None
    for i, x in enumerate(segmentation):
        x["trap"] = [i] * len(x["cell_label"])
        x["mother_assign_dynamic"] = np.array(x["mother_assign"])[
            np.array(x["cell_label"], dtype=int) - 1
        ]
    # Merge into a dictionary of lists, by column
    merged = {
        k: list(itertools.chain.from_iterable(res[k] for res in segmentation))
        for k in segmentation[0].keys()
    }
    # Special case for mother_assign
    # merged["mother_assign_dynamic"] = [merged["mother_assign"]]
    if "mother_assign" in merged:
        del merged["mother_assign"]
    #     mother_assign = [x["mother_assign"] for x in segmentation]
    # Check that the lists are all of the same length (in case of errors in
    # BABY)
    n_cells = min([len(v) for v in merged.values()])
    merged = {k: v[:n_cells] for k, v in merged.items()}
    merged["timepoint"] = [tp] * n_cells
    # merged["mother_assign"] = mother_assign
    return merged


class BabyParameters(ParametersABC):
    def __init__(
        self,
        model_config: t.Dict[
            str, t.Union[str, int, float, bool, t.List[t.Union[int, float]]]
        ],
        tracker_params,
        clogging_thresh,
        min_bud_tps,
        isbud_thresh,
        session,
        graph,
        print_info,
        suppress_errors,
        error_dump_dir,
        tf_version: int,
    ):
        self.model_config = model_config
        self.tracker_params = tracker_params
        self.clogging_thresh = clogging_thresh
        self.min_bud_tps = min_bud_tps
        self.isbud_thresh = isbud_thresh
        self.session = session
        self.graph = graph
        self.print_info = print_info
        self.suppress_errors = suppress_errors
        self.error_dump_dir = error_dump_dir
        self.tf_version = tf_version

    @classmethod
    def default(cls, **kwargs):
        """kwargs passes values to the model chooser"""
        return cls(
            model_config=choose_model_from_params(**kwargs),
            tracker_params=dict(ctrack_params=dict(), budtrack_params=dict()),
            clogging_thresh=1,
            min_bud_tps=3,
            isbud_thresh=0.5,
            session=None,
            graph=None,
            print_info=False,
            suppress_errors=False,
            error_dump_dir=None,
            tf_version=2,
        )

    def update_baby_modelset(self, path: t.Union[str, Path, t.Dict[str, str]]):
        """
        Replace default BABY model and flattener with another one from a folder outputted
        by our standard retraining script.
        """

        if isinstance(path, dict):
            weights_flattener = {k: Path(v) for k, v in path.items()}
        else:
            weights_dir = Path(path)
            weights_flattener = {
                "flattener_file": weights_dir.parent / "flattener.json",
                "morph_model_file": weights_dir / "weights.h5",
            }

        self.update("model_config", weights_flattener)


class BabyRunner(StepABC):
    """A BabyRunner object for cell segmentation.

    Does segmentation one time point at a time."""

    def __init__(self, tiler, parameters=None, **kwargs):
        self.tiler = tiler
        # self.model_config = modelsets()[choose_model_from_params(**kwargs)]
        self.model_config = (
            choose_model_from_params(**kwargs)
            if parameters is None
            else parameters.model_config
        )

        tiler_z = self.tiler.shape[-3]
        model_name = self.model_config["flattener_file"]
        if tiler_z != 5:
            assert (
                f"{tiler_z}z" in model_name
            ), f"Tiler z-stack ({tiler_z}) and Model shape ({model_name}) do not match "

        self.brain = BabyBrain(**self.model_config)
        self.crawler = BabyCrawler(self.brain)
        self.bf_channel = self.tiler.ref_channel_index

    @classmethod
    def from_tiler(cls, parameters: BabyParameters, tiler):
        return cls(tiler, parameters)

    def get_data(self, tp):
        # Swap axes x and z, probably shouldn't swap, just move z
        return (
            self.tiler.get_tp_data(tp, self.bf_channel)
            .swapaxes(1, 3)
            .swapaxes(1, 2)
        )

    def _run_tp(self, tp, with_edgemasks=True, assign_mothers=True, **kwargs):
        """Simulating processing time with sleep"""
        # Access the image
        t = perf_counter()
        img = self.get_data(tp)
        segmentation = self.crawler.step(
            img,
            with_edgemasks=with_edgemasks,
            assign_mothers=assign_mothers,
            **kwargs,
        )
        return format_segmentation(segmentation, tp)


def choose_model_from_params(
    modelset_filter=None,
    camera="prime95b",
    channel="brightfield",
    zoom="60x",
    n_stacks="5z",
):
    """
    Define which model to query from the server based on a set of parameters.

    Parameters
    ----------
    modelset_filter: str
                    A regex filter to apply on the models to start.
    camera: str
            The camera used in the experiment (case insensitive).
    channel:str
            The channel used for segmentation (case insensitive).
    zoom: str
          The zoom on the channel.
    n_stacks: str
              The number of z_stacks to use in segmentation

    Returns
    -------
    model_name : str
    """
    valid_models = list(modelsets().keys())

    # Apply modelset filter if specified
    if modelset_filter is not None:
        msf_regex = re.compile(modelset_filter)
        valid_models = filter(msf_regex.search, valid_models)

    # Apply parameter filters if specified
    params = [
        str(x) if x is not None else ".+"
        for x in [camera.lower(), channel.lower(), zoom, n_stacks]
    ]
    params_re = re.compile("^" + "_".join(params) + "$")
    valid_models = list(filter(params_re.search, valid_models))
    # Check that there are valid models
    if len(valid_models) == 0:
        raise KeyError(
            "No model sets found matching {}".format(", ".join(params))
        )
    # Pick the first model
    return modelsets()[valid_models[0]]



#!/usr/bin/env python3
import re
import typing as t

import numpy as np
import pandas as pd

from agora.io.bridge import groupsort
from itertools import groupby


def mb_array_to_dict(mb_array: np.ndarray):
    """
    Convert a lineage ndarray (trap, mother_id, daughter_id)
    into a dictionary of lists ( mother_id ->[daughters_ids] )
    """
    return {
        (trap, mo): [(trap, d[0]) for d in daughters]
        for trap, mo_da in groupsort(mb_array).items()
        for mo, daughters in groupsort(mo_da).items()
    }




import typing as t
from types import FunctionType
from inspect import getfullargspec, getmembers, isfunction, isbuiltin

import bottleneck as bn

from extraction.core.functions import cell, trap
from extraction.core.functions.custom import localisation
from extraction.core.functions.distributors import trap_apply
from extraction.core.functions.math_utils import div0

"""
Load functions for analysing cells and their background.
Note that inspect.getmembers returns a list of function names and functions,
and inspect.getfullargspec returns a function's arguments.
"""


def load_cellfuns_core():
    """Load functions from the cell module and return as a dict."""
    return {
        f[0]: f[1]
        for f in getmembers(cell)
        if isfunction(f[1])
        and f[1].__module__.startswith("extraction.core.functions")
    }


def load_custom_args() -> t.Tuple[
    (t.Dict[str, t.Callable], t.Dict[str, t.List[str]])
]:
    """
    Load custom functions from the localisation module.

    Return the functions and any additional arguments other
    than cell_mask and trap_image as dictionaries.
    """
    # load functions from module
    funs = {
        f[0]: f[1]
        for f in getmembers(localisation)
        if isfunction(f[1])
        and f[1].__module__.startswith("extraction.core.functions")
    }
    # load additional arguments if cell_mask and trap_image are arguments
    args = {
        k: getfullargspec(v).args[2:]
        for k, v in funs.items()
        if set(["cell_mask", "trap_image"]).intersection(
            getfullargspec(v).args
        )
    }
    # return dictionaries of functions and of arguments
    return (
        {k: funs[k] for k in args.keys()},
        {k: v for k, v in args.items() if v},
    )


def load_cellfuns():
    """
    Create a dict of core functions for use on cell_masks.

    The core functions only work on a single mask.
    """
    # create dict of the core functions from cell.py - these functions apply to a single mask
    cell_funs = load_cellfuns_core()
    # create a dict of functions that apply the core functions to an array of cell_masks
    CELLFUNS = {}
    for f_name, f in cell_funs.items():
        if isfunction(f):

            def tmp(f):
                args = getfullargspec(f).args
                if len(args) == 1:
                    # function that applies f to m, an array of masks
                    return lambda m, _: trap_apply(f, m)
                else:
                    # function that applies f to m and img, the trap_image
                    return lambda m, img: trap_apply(f, m, img)

            CELLFUNS[f_name] = tmp(f)
    return CELLFUNS


def load_trapfuns():
    """Load functions that are applied to an entire tile."""
    TRAPFUNS = {
        f[0]: f[1]
        for f in getmembers(trap)
        if isfunction(f[1])
        and f[1].__module__.startswith("extraction.core.functions")
    }
    return TRAPFUNS


def load_funs():
    """Combine all automatically loaded functions."""
    CELLFUNS = load_cellfuns()
    TRAPFUNS = load_trapfuns()
    # return dict of cell funs, dict of trap funs, and dict of both
    return CELLFUNS, TRAPFUNS, {**TRAPFUNS, **CELLFUNS}


def load_redfuns(
    additional_reducers: t.Optional[
        t.Union[t.Dict[str, t.Callable], t.Callable]
    ] = None,
) -> t.Dict[str, t.Callable]:
    """
    Load functions to reduce a multidimensional image by one dimension.

    Parameters
    ----------
    additional_reducers: function or a dict of functions (optional)
        Functions to perform the reduction.
    """
    RED_FUNS = {
        "max": bn.nanmax,
        "mean": bn.nanmean,
        "median": bn.nanmedian,
        "div0": div0,
        "add": bn.nansum,
        "None": None,
    }
    if additional_reducers is not None:
        if isinstance(additional_reducers, FunctionType):
            additional_reducers = [
                (additional_reducers.__name__, additional_reducers)
            ]
        RED_FUNS.update(additional_reducers)
    return RED_FUNS



#!/usr/bin/env python3

import numpy as np
import pandas as pd
from agora.abc import ParametersABC
from pycatch22 import catch22_all
from sklearn import decomposition

from postprocessor.core.abc import PostProcessABC


class catch22Parameters(ParametersABC):
    """
    Parameters
        :min_len: Prefilter to account only for long-signal cells
    """

    _defaults = {
        "min_len": 0.8,
        "n_components": None,
    }


class catch22(PostProcessABC):
    """
    catch22 class. It produces 22 normalised features for each time lapse in the signal (using the catch22 library.)
    """

    def __init__(self, parameters: catch22Parameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        thresh = (
            self.min_len
            if isinstance(self.min_len, int)
            else signal.shape[1] * self.min_len
        )
        adf = signal.loc[signal.notna().sum(axis=1) > thresh]
        catches = [
            catch22_all(adf.iloc[i, :].dropna().values)
            for i in range(len(adf))
        ]

        norm = pd.DataFrame(
            [catches[j]["values"] for j in range(len(catches))],
            index=adf.index,
            columns=catches[0]["names"],
        )

        return norm



import numpy as np


def div0(array, fill=0, axis=-1):
    """
    Divide array a by array b.

    If the result is a scalar and infinite, return fill.

    If the result contain elements that are infinite, replace these elements with fill.

    Parameters
    ----------
    a: array
    b: array
    fill: float
    **kwargs: kwargs
    """
    assert array.shape[axis] == 2, f"Array has the wrong shape in axis {axis}"
    slices_0, slices_1 = [[slice(None)] * len(array.shape)] * 2
    slices_0[axis] = 0
    slices_1[axis] = 1
    with np.errstate(divide="ignore", invalid="ignore"):
        c = np.true_divide(
            array[tuple(slices_0)],
            array[tuple(slices_1)],
        )
    if np.isscalar(c):
        return c if np.isfinite(c) else fill
    else:
        c[~np.isfinite(c)] = fill
        return c



"""
Asynchronous annotation (in one thread). Used as a base to build threading-based annotation.
Currently only works on UNIX-like systems due to using "/" to split addresses.

Usage example

From python
$ python annotator.py --image_path path/to/folder/with/h5files --results_path path/to/folder/with/images/zarr --pos position_name --ncells max_n_to_annotate

As executable (installed via poetry)
$ annotator.py --image_path path/to/folder/with/h5files --results_path path/to/folder/with/images/zarr --pos position_name --ncells max_n_to_annotate

During annotation:
- Assign a (binary) label by typing '1' or '2'.
- Type 'u' to undo.
- Type 's' to skip.
- Type 'q' to quit.

File will be saved in: ./YYYY-MM-DD_annotation/annotation.csv, where YYYY-MM-DD is the current date.

"""
import argparse
import logging
import typing as t

from copy import copy
from datetime import datetime
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import readchar
import trio
from agora.utils.cast import _str_to_int
from aliby.utils.vis_tools import _sample_n_tiles_masks
from aliby.utils.plot import stretch

# Remove logging warnings
logging.getLogger("aliby").setLevel(40)

# Defaults

essential = {"image_path": "zarr", "results_path": "h5"}
param_values = dict(
    out_dir=f"./{datetime.today().strftime('%Y_%m_%d')}_annotation/",
    pos=None,
    ncells=100,
    min_tp=100,
    max_tp=150,
    seed=0,
)

annot_filename = "annotation.csv"

# Parsing
parser = argparse.ArgumentParser(
    prog="aliby-annot-binary",
    description="Annotate cells in a binary manner",
)
for i, arg in enumerate((*essential, *param_values)):
    parser.add_argument(
        f"--{arg}",
        action="store",
        default=param_values.get(arg),
        required=i < len(essential),
    )

args = parser.parse_args()

for i, k in enumerate((*essential, *param_values.keys())):
    # Assign essential values as-is
    if i < len(essential):
        param_values[k] = getattr(args, k)

    # Fill additional values
    if passed_value := getattr(args, k):
        param_values[k] = passed_value
        try:
            param_values[k] = _str_to_int(passed_value)
        except Exception as exc:
            pass

for k, suffix in essential.items():  # Autocomplete if fullpath not provided
    if not str(param_values[k]).endswith(suffix):
        param_values[k] = (
            Path(param_values[k]) / f"{ param_values['pos'] }.{suffix}"
        )


# Functions
async def generate_image(stack, skip: bool = False):
    await trio.sleep(1)
    result = np.random.randint(100, size=(10, 10))
    stack.append(result)


async def draw(data, drawing):
    if len(drawing) > 1:
        for ax, img in zip(drawing, data):
            if np.isnan(img).sum():  # Stretch masked channel
                img = stretch(img)
            ax.set_data(img)
    else:
        drawing.set_data(data)
    plt.draw()
    plt.pause(0.1)


def annotate_image(current_key=None, valid_values: t.Tuple[int] = (1, 2)):
    # Show image to annotate
    while current_key is None or current_key not in valid_values:
        if current_key is not None:
            print(
                f"Invalid value. Please try with valid values {valid_values}"
            )

        if (current_key := readchar.readkey()) in "qsu":
            # if (current_key := input()) in "qsu":
            break

        current_key = _parse_input(current_key, valid_values)

    return current_key


async def generate_image(
    generator,
    location_stack: t.List[t.Tuple[np.ndarray, t.Tuple[int, int, int]]],
):
    new_location_image = next(generator)
    location_stack.append((new_location_image[0], new_location_image[1]))


def _parse_input(value: str, valid_values: t.Tuple[int]):
    try:
        return int(value)
    except:
        print(
            f"Non-parsable value. Please try again with valid values {valid_values}"
        )
        return None


def write_annotation(
    experiment_position: str,
    out_dir: Path,
    annotation: str,
    location_stack: t.Tuple[t.Tuple[int, int, int], np.ndarray],
):
    location, stack = location_stack
    unique_location = list(map(str, (*experiment_position, *location)))

    write_into_file(
        out_dir / annot_filename,
        ",".join((*unique_location, str(annotation))) + "\n",
    )
    bg_zero = copy(stack[1])
    bg_zero[np.isnan(bg_zero)] = 0
    tosave = np.stack((stack[0], bg_zero.astype(int)))
    # np.savez(out_dir / f"{'_'.join( unique_location )}.npz", tosave)
    np.save(out_dir / f"{'.'.join( unique_location )}.npy", tosave)


def write_into_file(file_path: str, line: str):
    with open(file_path, "a") as f:
        f.write(str(line))


async def annotate_images(
    image_path, results_path, out_dir, ncells, seed, interval
):

    preemptive_cache = 3

    location_stack = []
    out_dir = Path(out_dir)
    out_annot_file = str(out_dir / annot_filename)

    generator = _sample_n_tiles_masks(
        image_path, results_path, ncells, seed=seed, interval=interval
    )

    # Fetch a few positions preemtively
    async with trio.open_nursery() as nursery:
        for _ in range(preemptive_cache):
            nursery.start_soon(generate_image, generator, location_stack)

        print("parent: waiting for first annotations.")

        _, ax = plt.subplots(figsize=(10, 8))
        while not location_stack:  # Wait until first image is loaded
            await trio.sleep(0.1)

        from aliby.utils.plot import plot_overlay

        # drawing = ax.imshow(location_stack[0][1])
        axes = plot_overlay(*location_stack[0][1], ax=ax.axes)
        plt.show(block=False)
        plt.draw()
        plt.pause(0.5)  # May be adjusted based on display speed

    try:
        out_dir.mkdir(parents=True)
    except:
        pass
    if not Path(out_annot_file).exists():
        write_into_file(
            out_annot_file,
            ",".join(
                (
                    "experiment",
                    "position",
                    "tile",
                    "cell_label",
                    "tp",
                    "annotation",
                )
            )
            + "\n",
        )

    # Loop until n_max or quit
    for i in range(1, ncells - preemptive_cache + 1):
        # Wait for input
        print("Enter a key")
        annotation = str(annotate_image())

        if annotation == "q":
            break
        elif annotation == "s":
            print("Skipping...")
            # continue
        elif annotation == "u":
            i -= 1
        elif isinstance(_str_to_int(annotation), int):
            write_annotation(
                str(results_path).split(".")[0].split("/")[-2:],
                out_dir,
                annotation,
                location_stack[i],
            )

        print(location_stack[i][0])
        # Append into annotations file
        async with trio.open_nursery() as nursery:
            nursery.start_soon(generate_image, generator, location_stack)
            nursery.start_soon(draw, location_stack[i][1], axes)

    print("Annotation done!")


# if __name__ == "__main__":
def annotate():
    if any([param_values.get(k) is None for k in ("min_tp", "max_tp")]):
        interval = None
    else:
        interval = (param_values["min_tp"], param_values["max_tp"])

    print(param_values)
    trio.run(
        annotate_images,
        param_values["image_path"],
        param_values["results_path"],
        param_values["out_dir"],
        param_values["ncells"],
        param_values["seed"],
        interval,
    )



"""Routines for analysing post-processed data that don't follow the parameters-processes structure.

Routines for analysing post-processed data that don't follow the
parameters-processes structure.

Currently, these consist of plotting routines.  There is one module for each
plotting routine.  Each module consists of two components and is structured as
follows:
1. An internal class.
    The class defines the parameters and defines additional class attributes to
    help with plotting.  The class also has one method (`plot`) that takes a
    `matplotlib.Axes` object as an argument.  This method draws the plot on the
    `Axes` object.
2. A plotting function.
    The user accesses this function.  This function defines the default
    parameters in its arguments.  Within the function, a 'plotter' object is
    defined using the internal class and then the function draws the plot on a
    specified `matplotlib.Axes` object.

This structure follows that of plotting routines in `seaborn`
(https://github.com/mwaskom/seaborn), a Python visualisation library that is
based on `matplotlib`.
"""



"""
Neural network initialisation.
"""
from pathlib import Path
from time import perf_counter

import numpy as np
import tensorflow as tf
from agora.io.writer import DynamicWriter


def initialise_tf(version):
    # Initialise tensorflow
    if version == 1:
        core_config = tf.ConfigProto()
        core_config.gpu_options.allow_growth = True
        session = tf.Session(config=core_config)
        return session
    # TODO this only works for TF2
    if version == 2:
        gpus = tf.config.experimental.list_physical_devices("GPU")
        if gpus:
            for gpu in gpus:
                tf.config.experimental.set_memory_growth(gpu, True)
            logical_gpus = tf.config.experimental.list_logical_devices("GPU")
            print(
                len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs"
            )
        return None


def timer(func, *args, **kwargs):
    start = perf_counter()
    result = func(*args, **kwargs)
    print(f"Function {func.__name__}: {perf_counter() - start}s")
    return result


################## CUSTOM OBJECTS ##################################
class ModelPredictor:
    """Generic object that takes a NN and returns the prediction.

    Use for predicting fluorescence/other from bright field.
    This does not do instance segmentations of anything.
    """

    def __init__(self, tiler, model, name):
        self.tiler = tiler
        self.model = model
        self.name = name

    def get_data(self, tp):
        # Change axes to X,Y,Z rather than Z,Y,X
        return (
            self.tiler.get_tp_data(tp, self.bf_channel)
            .swapaxes(1, 3)
            .swapaxes(1, 2)
        )

    def format_result(self, result, tp):
        return {self.name: result, "timepoints": [tp] * len(result)}

    def run_tp(self, tp):
        """Simulating processing time with sleep"""
        # Access the image
        segmentation = self.model.predict(self.get_data(tp))
        return self._format_result(segmentation, tp)


class ModelPredictorWriter(DynamicWriter):
    def __init__(self, file, name, shape, dtype):
        super.__init__(file)
        self.datatypes = {
            name: (shape, dtype),
            "timepoint": ((None,), np.uint16),
        }
        self.group = f"{self.name}_info"



#!/usr/bin/env python3

import re
from itertools import product
from pydoc import locate

from agora.abc import ParametersABC, ProcessABC


class PostProcessABC(ProcessABC):
    """
    Extend ProcessABC to add as_function, allowing for all PostProcesses to be called as functions
    almost directly.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def as_function(cls, data, *extra_data, **kwargs):
        # FIXME can this be a __call__ method instead?
        # Find the parameter's default
        parameters = cls.default_parameters(**kwargs)
        return cls(parameters=parameters).run(data, *extra_data)

    @classmethod
    def default_parameters(cls, *args, **kwargs):
        return get_parameters(cls.__name__).default(*args, **kwargs)


def get_process(process, suffix="") -> PostProcessABC or ParametersABC or None:
    """
    Dynamically import a process class from the available process locations.
    Assumes process filename and class name are the same

    Processes return the same shape as their input.
    MultiSignal either take or return multiple datasets (or both).
    Reshapers return a different shape for processes: Merger and Picker belong here.

    suffix : str
        Name of suffix, generally "" (empty) or "Parameters".
    """
    base_location = "postprocessor.core"
    possible_locations = ("processes", "multisignal", "reshapers")
    valid_syntaxes = (
        _to_snake_case(process),
        _to_pascal_case(_to_snake_case(process)),
    )

    found = None
    for possible_location, process_syntax in product(
        possible_locations, valid_syntaxes
    ):

        location = f"{base_location}.{possible_location}.{_to_snake_case(process)}.{process_syntax}{suffix}"
        found = locate(location)
        if found is not None:
            break
    else:

        raise Exception(
            f"{process} not found in locations {possible_locations} at {base_location}"
        )
    return found


def get_parameters(process: str) -> ParametersABC:
    """
    Dynamically import parameters from the 'processes' folder.
    Assumes parameter is the same name as the file with 'Parameters' added at the end.
    """
    return get_process(process, suffix="Parameters")


def _to_pascal_case(snake_str: str) -> str:
    # Convert a snake_case string to PascalCase.
    # Based on https://stackoverflow.com/a/19053800
    components = snake_str.split("_")
    return "".join(x.title() for x in components)


def _to_snake_case(pascal_str: str) -> str:
    # Convert a snake_case string to PascalCase.
    # Based on https://stackoverflow.com/a/12867228
    return re.sub("(?!^)([A-Z]+)", r"_\1", pascal_str).lower()



"""
Functions to process, filter and merge tracks.
"""

# from collections import Counter

import typing as t
from copy import copy
from typing import List, Union

import more_itertools as mit
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from utils_find_1st import cmp_larger, find_1st

from postprocessor.core.processes.savgol import non_uniform_savgol


def load_test_dset():
    """Load development dataset to test functions."""
    return pd.DataFrame(
        {
            ("a", 1, 1): [2, 5, np.nan, 6, 8] + [np.nan] * 5,
            ("a", 1, 2): list(range(2, 12)),
            ("a", 1, 3): [np.nan] * 8 + [6, 7],
            ("a", 1, 4): [np.nan] * 5 + [9, 12, 10, 14, 18],
        },
        index=range(1, 11),
    ).T


def max_ntps(track: pd.Series) -> int:
    """Get number of time points."""
    indices = np.where(track.notna())
    return np.max(indices) - np.min(indices)


def max_nonstop_ntps(track: pd.Series) -> int:
    nona_tracks = track.notna()
    consecutive_nonas_grouped = [
        len(list(x))
        for x in mit.consecutive_groups(np.flatnonzero(nona_tracks))
    ]
    return max(consecutive_nonas_grouped)


def get_tracks_ntps(tracks: pd.DataFrame) -> pd.Series:
    return tracks.apply(max_ntps, axis=1)


def get_avg_gr(track: pd.Series) -> int:
    """
    Get average growth rate for a track.

    :param tracks: Series with volume and timepoints as indices
    """
    ntps = max_ntps(track)
    vals = track.dropna().values
    gr = (vals[-1] - vals[0]) / ntps
    return gr


def get_avg_grs(tracks: pd.DataFrame) -> pd.DataFrame:
    """
    Get average growth rate for a group of tracks

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    """
    return tracks.apply(get_avg_gr, axis=1)


def clean_tracks(
    tracks, min_len: int = 15, min_gr: float = 1.0
) -> pd.DataFrame:
    """
    Clean small non-growing tracks and return the reduced dataframe

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param min_len: int number of timepoints cells must have not to be removed
    :param min_gr: float Minimum mean growth rate to assume an outline is growing
    """
    ntps = get_tracks_ntps(tracks)
    grs = get_avg_grs(tracks)
    growing_long_tracks = tracks.loc[(ntps >= min_len) & (grs > min_gr)]
    return growing_long_tracks


def merge_tracks(
    tracks, drop=False, **kwargs
) -> t.Tuple[pd.DataFrame, t.Collection]:
    """
    Join tracks that are contiguous and within a volume threshold of each other

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param kwargs: args passed to get_joinable

    returns

    :joint_tracks: (m x n) Dataframe where rows are cell tracks and
        columns are timepoints. Merged tracks are still present but filled
        with np.nans.
    """

    # calculate tracks that can be merged until no more traps can be merged
    joinable_pairs = get_joinable(tracks, **kwargs)
    if joinable_pairs:
        tracks = join_tracks(tracks, joinable_pairs, drop=drop)
    return (tracks, joinable_pairs)


def get_joint_ids(merging_seqs) -> dict:
    """
    Convert a series of merges into a dictionary where
    the key is the cell_id of destination and the value a list
    of the other track ids that were merged into the key

    :param merging_seqs: list of tuples of indices indicating the
    sequence of merging events. It is important for this to be in sequential order

    How it works:

    The order of merging matters for naming, always the leftmost track will keep the id

    For example, having tracks (a, b, c, d) and the iterations of merge events:

    0 a b c d
    1 a b cd
    2 ab cd
    3 abcd

    We should get:

    output {a:a, b:a, c:a, d:a}

    """
    if not merging_seqs:
        return {}
    targets, origins = list(zip(*merging_seqs))
    static_tracks = set(targets).difference(origins)
    joint = {track_id: track_id for track_id in static_tracks}
    for target, origin in merging_seqs:
        joint[origin] = target
    moved_target = [
        k for k, v in joint.items() if joint[v] != v and v in joint.values()
    ]
    for orig in moved_target:
        joint[orig] = rec_bottom(joint, orig)
    return {
        k: v for k, v in joint.items() if k != v
    }  # remove ids that point to themselves


def rec_bottom(d, k):
    if d[k] == k:
        return k
    else:
        return rec_bottom(d, d[k])


def join_tracks(tracks, joinable_pairs, drop=True) -> pd.DataFrame:
    """
    Join pairs of tracks from later tps towards the start.

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints

    returns (copy)

    :param joint_tracks: (m x n) Dataframe where rows are cell tracks and
        columns are timepoints. Merged tracks are still present but filled
        with np.nans.
    :param drop: bool indicating whether or not to drop moved rows

    """
    tmp = copy(tracks)
    for target, source in joinable_pairs:
        tmp.loc[target] = join_track_pair(tmp.loc[target], tmp.loc[source])
        if drop:
            tmp = tmp.drop(source)
    return tmp


def join_track_pair(target, source):
    tgt_copy = copy(target)
    end = find_1st(target.values[::-1], 0, cmp_larger)
    tgt_copy.iloc[-end:] = source.iloc[-end:].values
    return tgt_copy


def get_joinable(tracks, smooth=False, tol=0.1, window=5, degree=3) -> dict:
    """
    Get the pair of track (without repeats) that have a smaller error than the
    tolerance. If there is a track that can be assigned to two or more other
    ones, choose the one with lowest error.

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param tol: float or int threshold of average (prediction error/std) necessary
        to consider two tracks the same. If float is fraction of first track,
        if int it is absolute units.
    :param window: int value of window used for savgol_filter
    :param degree: int value of polynomial degree passed to savgol_filter

    """
    tracks = tracks.loc[tracks.notna().sum(axis=1) > 2]

    # Commented because we are not smoothing in this step yet
    # candict = {k:v for d in contig.values for k,v in d.items()}

    # smooth all relevant tracks

    if smooth:  # Apply savgol filter TODO fix nans affecting edge placing
        clean = clean_tracks(
            tracks, min_len=window + 1, min_gr=0.9
        )  # get useful tracks

        def savgol_on_srs(x):
            return non_uniform_savgol(x.index, x.values, window, degree)

        contig = clean.groupby(["trap"]).apply(get_contiguous_pairs)
        contig = contig.loc[contig.apply(len) > 0]
        flat = set([k for v in contig.values for i in v for j in i for k in j])
        smoothed_tracks = clean.loc[flat].apply(savgol_on_srs, 1)
    else:
        contig = tracks.groupby(["trap"]).apply(get_contiguous_pairs)
        contig = contig.loc[contig.apply(len) > 0]
        flat = set([k for v in contig.values for i in v for j in i for k in j])
        smoothed_tracks = tracks.loc[flat].apply(
            lambda x: np.array(x.values), axis=1
        )

    # fetch edges from ids TODO (IF necessary, here we can compare growth rates)
    def idx_to_edge(preposts):
        return [
            (
                [get_val(smoothed_tracks.loc[pre], -1) for pre in pres],
                [get_val(smoothed_tracks.loc[post], 0) for post in posts],
            )
            for pres, posts in preposts
        ]

    # idx_to_means = lambda preposts: [
    #     (
    #         [get_means(smoothed_tracks.loc[pre], -window) for pre in pres],
    #         [get_means(smoothed_tracks.loc[post], window) for post in posts],
    #     )
    #     for pres, posts in preposts
    # ]

    def idx_to_pred(preposts):
        result = []
        for pres, posts in preposts:
            pre_res = []
            for pre in pres:
                y = get_last_i(smoothed_tracks.loc[pre], -window)
                pre_res.append(
                    np.poly1d(np.polyfit(range(len(y)), y, 1))(len(y) + 1),
                )
            pos_res = [
                get_means(smoothed_tracks.loc[post], window) for post in posts
            ]
            result.append([pre_res, pos_res])

        return result

    edges = contig.apply(idx_to_edge)  # Raw edges
    # edges_mean = contig.apply(idx_to_means)  # Mean of both
    pre_pred = contig.apply(idx_to_pred)  # Prediction of pre and mean of post

    # edges_dMetric = edges.apply(get_dMetric_wrap, tol=tol)
    # edges_dMetric_mean = edges_mean.apply(get_dMetric_wrap, tol=tol)
    edges_dMetric_pred = pre_pred.apply(get_dMetric_wrap, tol=tol)

    # combined_dMetric = pd.Series(
    #     [
    #         [np.nanmin((a, b), axis=0) for a, b in zip(x, y)]
    #         for x, y in zip(edges_dMetric, edges_dMetric_mean)
    #     ],
    #     index=edges_dMetric.index,
    # )
    # closest_pairs = combined_dMetric.apply(get_vec_closest_pairs, tol=tol)
    solutions = []
    # for (i, dMetrics), edgeset in zip(combined_dMetric.items(), edges):
    for (i, dMetrics), edgeset in zip(edges_dMetric_pred.items(), edges):
        solutions.append(solve_matrices_wrap(dMetrics, edgeset, tol=tol))

    closest_pairs = pd.Series(
        solutions,
        index=edges_dMetric_pred.index,
    )

    # match local with global ids
    joinable_ids = [
        localid_to_idx(closest_pairs.loc[i], contig.loc[i])
        for i in closest_pairs.index
    ]

    return [pair for pairset in joinable_ids for pair in pairset]


def get_val(x, n):
    return x[~np.isnan(x)][n] if len(x[~np.isnan(x)]) else np.nan


def get_means(x, i):
    if not len(x[~np.isnan(x)]):
        return np.nan
    if i > 0:
        v = x[~np.isnan(x)][:i]
    else:
        v = x[~np.isnan(x)][i:]
    return np.nanmean(v)


def get_last_i(x, i):
    if not len(x[~np.isnan(x)]):
        return np.nan
    if i > 0:
        v = x[~np.isnan(x)][:i]
    else:
        v = x[~np.isnan(x)][i:]
    return v


def localid_to_idx(local_ids, contig_trap):
    """
    Fetch the original ids from a nested list with joinable local_ids.

    input
    :param local_ids: list of list of pairs with cell ids to be joint
    :param local_ids: list of list of pairs with corresponding cell ids

    return
    list of pairs with (experiment-level) ids to be joint
    """
    lin_pairs = []
    for i, pairs in enumerate(local_ids):
        if len(pairs):
            for left, right in pairs:
                lin_pairs.append(
                    (contig_trap[i][0][left], contig_trap[i][1][right])
                )
    return lin_pairs


def get_vec_closest_pairs(lst: List, **kwargs):
    return [get_closest_pairs(*sublist, **kwargs) for sublist in lst]


def get_dMetric_wrap(lst: List, **kwargs):
    return [get_dMetric(*sublist, **kwargs) for sublist in lst]


def solve_matrices_wrap(dMetric: List, edges: List, **kwargs):
    return [
        solve_matrices(mat, edgeset, **kwargs)
        for mat, edgeset in zip(dMetric, edges)
    ]


def get_dMetric(
    pre: List[float], post: List[float], tol: Union[float, int] = 1
):
    """Calculate a cost matrix

    input
    :param pre: list of floats with edges on left
    :param post: list of floats with edges on right
    :param tol: int or float if int metrics of tolerance, if float fraction

    returns
    :: list of indices corresponding to the best solutions for matrices

    """
    if len(pre) > len(post):
        dMetric = np.abs(np.subtract.outer(post, pre))
    else:
        dMetric = np.abs(np.subtract.outer(pre, post))
    dMetric[np.isnan(dMetric)] = (
        tol + 1 + np.nanmax(dMetric)
    )  # nans will be filtered
    return dMetric


def solve_matrices(
    dMetric: np.ndarray, prepost: List, tol: Union[float, int] = 1
):
    """
    Solve the distance matrices obtained in get_dMetric and/or merged from
    independent dMetric matrices.
    """
    ids = solve_matrix(dMetric)
    if not len(ids[0]):
        return []
    pre, post = prepost
    norm = (
        np.array(pre)[ids[len(pre) > len(post)]] if tol < 1 else 1
    )  # relative or absolute tol
    result = dMetric[ids] / norm
    ids = ids if len(pre) < len(post) else ids[::-1]
    return [idx for idx, res in zip(zip(*ids), result) if res <= tol]


def get_closest_pairs(
    pre: List[float], post: List[float], tol: Union[float, int] = 1
):
    """
    Calculate a cost matrix for the Hungarian algorithm to pick the best set of
    options.

    input
    :param pre: list of floats with edges on left
    :param post: list of floats with edges on right
    :param tol: int or float if int metrics of tolerance, if float fraction

    returns
    :: list of indices corresponding to the best solutions for matrices

    """
    dMetric = get_dMetric(pre, post, tol)
    return solve_matrices(dMetric, pre, post, tol)


def solve_matrix(dMetric):
    """
    Solve cost matrix focusing on getting the smallest cost at each iteration.

    input
    :param dMetric: np.array cost matrix

    returns
    tuple of np.arrays indicating picks with lowest individual value
    """
    glob_is = []
    glob_js = []
    if (~np.isnan(dMetric)).any():
        tmp = copy(dMetric)
        std = sorted(tmp[~np.isnan(tmp)])
        while (~np.isnan(std)).any():
            v = std[0]
            i_s, j_s = np.where(tmp == v)
            i = i_s[0]
            j = j_s[0]
            tmp[i, :] += np.nan
            tmp[:, j] += np.nan
            glob_is.append(i)
            glob_js.append(j)
            std = sorted(tmp[~np.isnan(tmp)])
    return (np.array(glob_is), np.array(glob_js))


def plot_joinable(tracks, joinable_pairs):
    """
    Convenience plotting function for debugging and data vis
    """
    nx = 8
    ny = 8
    _, axes = plt.subplots(nx, ny)
    for i in range(nx):
        for j in range(ny):
            if i * ny + j < len(joinable_pairs):
                ax = axes[i, j]
                pre, post = joinable_pairs[i * ny + j]
                pre_srs = tracks.loc[pre].dropna()
                post_srs = tracks.loc[post].dropna()
                ax.plot(pre_srs.index, pre_srs.values, "b")
                # try:
                #     totrange = np.arange(pre_srs.index[0],post_srs.index[-1])
                #     ax.plot(totrange, interpolate(pre_srs, totrange), 'r-')
                # except:
                #     pass
                ax.plot(post_srs.index, post_srs.values, "g")
    plt.show()


def get_contiguous_pairs(tracks: pd.DataFrame) -> list:
    """
    Get all pair of contiguous track ids from a tracks dataframe.

    :param tracks: (m x n) dataframe where rows are cell tracks and
        columns are timepoints
    :param min_dgr: float minimum difference in growth rate from
        the interpolation
    """
    mins, maxes = [
        tracks.notna().apply(np.where, axis=1).apply(fn)
        for fn in (np.min, np.max)
    ]
    mins_d = mins.groupby(mins).apply(lambda x: x.index.tolist())
    mins_d.index = mins_d.index - 1  # make indices equal
    # TODO add support for skipping time points
    maxes_d = maxes.groupby(maxes).apply(lambda x: x.index.tolist())
    common = sorted(
        set(mins_d.index).intersection(maxes_d.index), reverse=True
    )
    return [(maxes_d[t], mins_d[t]) for t in common]


# def fit_track(track: pd.Series, obj=None):
#     if obj is None:
#         obj = objective

#     x = track.dropna().index
#     y = track.dropna().values
#     popt, _ = curve_fit(obj, x, y)

#     return popt

# def interpolate(track, xs) -> list:
#     '''
#     Interpolate next timepoint from a track

#     :param track: pd.Series of volume growth over a time period
#     :param t: int timepoint to interpolate
#     '''
#     popt = fit_track(track)
#     # perr = np.sqrt(np.diag(pcov))
#     return objective(np.array(xs), *popt)


# def objective(x,a,b,c,d) -> float:
#     # return (a)/(1+b*np.exp(c*x))+d
#     return (((x+d)*a)/((x+d)+b))+c

# def cand_pairs_to_dict(candidates):
#     d={x:[] for x,_ in candidates}
#     for x,y in candidates:
#         d[x].append(y)
#     return d



#!/usr/bin/env python3

import re
import typing as t
from abc import ABC, abstractproperty
from collections import Counter
from functools import cached_property as property
from pathlib import Path
from typing import Dict, List, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from pathos.multiprocessing import Pool

from postprocessor.chainer import Chainer


class Grouper(ABC):
    """Base grouper class."""

    def __init__(self, dir: Union[str, Path]):
        """Find h5 files and load a chain for each one."""
        path = Path(dir)
        assert path.exists(), f"{str(dir)} does not exist"
        self.name = path.name
        self.files = list(path.glob("*.h5"))
        assert len(self.files), "No valid h5 files in dir"
        self.load_chains()

    def load_chains(self) -> None:
        """Load a chain for each position, or h5 file."""
        self.chainers = {f.name[:-3]: Chainer(f) for f in self.files}

    @property
    def fsignal(self) -> Chainer:
        """Get first chain."""
        return list(self.chainers.values())[0]

    @property
    def ntimepoints(self) -> int:
        """Find number of time points."""
        return max([s.ntimepoints for s in self.chainers.values()])

    @property
    def tintervals(self) -> float:
        """Find the maximum time interval for all chains."""
        tintervals = set([s.tinterval / 60 for s in self.chainers.values()])
        assert (
            len(tintervals) == 1
        ), "Not all chains have the same time interval"
        return max(tintervals)

    @property
    def available(self) -> t.Collection[str]:
        """Generate list of available signals in the first chain."""
        return self.fsignal.available

    @property
    def available_grouped(self) -> None:
        """Display available signals and the number of chains for each."""
        if not hasattr(self, "_available_grouped"):
            self._available_grouped = Counter(
                [x for s in self.chainers.values() for x in s.available]
            )
        for s, n in self._available_grouped.items():
            print(f"{s} - {n}")

    @property
    def datasets(self) -> None:
        """Print available data sets in the first chain."""
        return self.fsignal.datasets

    @abstractproperty
    def positions_groups(self):
        pass

    def concat_signal(
        self,
        path: str,
        pool: t.Optional[int] = None,
        mode: str = "retained",
        standard: t.Optional[bool] = False,
        **kwargs,
    ):
        """
        Concatenate data for one signal from different h5 files, with
        one h5 file per position, into a dataframe.

        Parameters
        ----------
        path : str
           Signal location within h5py file
        pool : int
           Number of threads used; if 0 or None only one core is used
        mode: str
        standard: boolean
        **kwargs : key, value pairings
           Named arguments for concat_ind_function

        Examples
        --------
        >>> record = grouper.concat_signal("extraction/GFP/max/median")
        """
        if path.startswith("/"):
            path = path.strip("/")
        good_chains = self.filter_chains(path)
        if standard:
            fn_pos = concat_standard
        else:
            fn_pos = concat_signal_ind
            kwargs["mode"] = mode
        records = self.pool_function(
            path=path,
            f=fn_pos,
            pool=pool,
            chainers=good_chains,
            **kwargs,
        )
        # check for errors
        errors = [
            k for kymo, k in zip(records, self.chainers.keys()) if kymo is None
        ]
        records = [record for record in records if record is not None]
        if len(errors):
            print("Warning: Positions contain errors {errors}")
        assert len(records), "All data sets contain errors"
        # combine into one dataframe
        concat = pd.concat(records, axis=0)
        if len(concat.index.names) > 4:
            # reorder levels in the multi-index dataframe when mother_label is present
            concat = concat.reorder_levels(
                ("group", "position", "trap", "cell_label", "mother_label")
            )
        concat_sorted = concat.sort_index()
        return concat_sorted

    def filter_chains(self, path: str) -> t.Dict[str, Chainer]:
        """Filter chains to those whose data is available in the h5 file."""
        good_chains = {
            k: v
            for k, v in self.chainers.items()
            if path in [*v.available, *v.common_chains]
        }
        nchains_dif = len(self.chainers) - len(good_chains)
        if nchains_dif:
            print(
                f"Grouper:Warning: {nchains_dif} chains do not contain"
                f" channel {path}"
            )
        assert len(
            good_chains
        ), f"No valid dataset to use. Valid datasets are {self.available}"
        return good_chains

    def pool_function(
        self,
        path: str,
        f: t.Callable,
        pool: t.Optional[int] = None,
        chainers: t.Dict[str, Chainer] = None,
        **kwargs,
    ):
        """Enable different threads for independent chains, particularly useful when aggregating multiple elements."""
        if pool is None:
            pass
        chainers = chainers or self.chainers
        if pool:
            with Pool(pool) as p:
                records = p.map(
                    lambda x: f(
                        path=path,
                        chainer=x[1],
                        group=self.positions_groups[x[0]],
                        position=x[0],
                        **kwargs,
                    ),
                    chainers.items(),
                )
        else:
            records = [
                f(
                    path=path,
                    chainer=chainer,
                    group=self.positions_groups[name],
                    position=name,
                    **kwargs,
                )
                for name, chainer in self.chainers.items()
            ]
        return records

    @property
    def nmembers(self) -> t.Dict[str, int]:
        """Get the number of positions belonging to each group."""
        return Counter(self.positions_groups.values())

    @property
    def ntiles(self):
        """Get total number of tiles per position (h5 file)."""
        for pos, s in self.chainers.items():
            with h5py.File(s.filename, "r") as f:
                print(pos, f["/trap_info/trap_locations"].shape[0])

    @property
    def ntiles_by_group(self) -> t.Dict[str, int]:
        """Get total number of tiles per group."""
        ntiles = {}
        for pos, s in self.chainers.items():
            with h5py.File(s.filename, "r") as f:
                ntiles[pos] = f["/trap_info/trap_locations"].shape[0]
        ntiles_by_group = {k: 0 for k in self.groups}
        for posname, vals in ntiles.items():
            ntiles_by_group[self.positions_groups[posname]] += vals
        return ntiles_by_group

    @property
    def tilelocs(self) -> t.Dict[str, np.ndarray]:
        """Get the locations of the tiles for each position as a dictionary."""
        d = {}
        for pos, s in self.chainers.items():
            with h5py.File(s.filename, "r") as f:
                d[pos] = f["/trap_info/trap_locations"][()]
        return d

    @property
    def groups(self) -> t.Tuple[str]:
        """Get groups, sorted alphabetically."""
        return tuple(sorted(set(self.positions_groups.values())))

    @property
    def positions(self) -> t.Tuple[str]:
        """Get positions, sorted alphabetically."""
        return tuple(sorted(set(self.positions_groups.keys())))

    def ncells(
        self,
        path="extraction/general/None/area",
        mode="retained",
        **kwargs,
    ) -> t.Dict[str, int]:
        """Get number of cells retained per position in base channel as a dictionary."""
        return (
            self.concat_signal(path=path, mode=mode, **kwargs)
            .groupby("group")
            .apply(len)
            .to_dict()
        )

    @property
    def nretained(self) -> t.Dict[str, int]:
        """Get number of cells retained per position in base channel as a dictionary."""
        return self.ncells()

    @property
    def channels(self):
        """Get unique channels for all chains as a set."""
        return set(
            [
                channel
                for chainer in self.chainers.values()
                for channel in chainer.channels
            ]
        )

    @property
    def stages_span(self):
        # FAILS on my example
        return self.fsignal.stages_span

    @property
    def max_span(self):
        # FAILS on my example
        return self.fsignal.max_span

    @property
    def stages(self):
        # FAILS on my example
        return self.fsignal.stages

    @property
    def tinterval(self):
        """Get interval between time points."""
        return self.fsignal.tinterval


class MetaGrouper(Grouper):
    """Group positions using metadata's 'group' number."""

    pass


class NameGrouper(Grouper):
    """Group a set of positions with a shorter version of the group's name."""

    def __init__(self, dir, name_inds=(0, -4)):
        """Define the indices to slice names."""
        super().__init__(dir=dir)
        self.name_inds = name_inds

    @property
    def positions_groups(self) -> t.Dict[str, str]:
        """Get a dictionary with the positions as keys and groups as items."""
        if not hasattr(self, "_positions_groups"):
            self._positions_groups = {}
            for name in self.chainers.keys():
                self._positions_groups[name] = name[
                    self.name_inds[0] : self.name_inds[1]
                ]
        return self._positions_groups


class phGrouper(NameGrouper):
    """Grouper for pH calibration experiments where all surveyed media pH values are within a single experiment."""

    def __init__(self, dir, name_inds=(3, 7)):
        """Initialise via NameGrouper."""
        super().__init__(dir=dir, name_inds=name_inds)

    def get_ph(self) -> None:
        """Find the pH from the group names and store as a dictionary."""
        self.ph = {gn: self.ph_from_group(gn) for gn in self.positions_groups}

    @staticmethod
    def ph_from_group(group_name: str) -> float:
        """Find the pH from the name of a group."""
        if group_name.startswith("ph_") or group_name.startswith("pH_"):
            group_name = group_name[3:]
        return float(group_name.replace("_", "."))

    def aggregate_multichains(self, signals: list) -> pd.DataFrame:
        """Get data from a list of signals and combine into one multi-index dataframe with 'media-pH' included."""
        aggregated = pd.concat(
            [
                self.concat_signal(signal, reduce_cols=np.nanmean)
                for signal in signals
            ],
            axis=1,
        )
        ph = pd.Series(
            [
                self.ph_from_group(
                    x[list(aggregated.index.names).index("group")]
                )
                for x in aggregated.index
            ],
            index=aggregated.index,
            name="media_pH",
        )
        aggregated = pd.concat((aggregated, ph), axis=1)
        return aggregated


def concat_standard(
    path: str,
    chainer: Chainer,
    group: str,
    position: t.Optional[str] = None,
    **kwargs,
) -> pd.DataFrame:
    combined = chainer.get(path, **kwargs).copy()
    combined["position"] = position
    combined["group"] = group
    combined.set_index(["group", "position"], inplace=True, append=True)
    combined.index = combined.index.reorder_levels(
        ("group", "position", "trap", "cell_label", "mother_label")
    )
    return combined


def concat_signal_ind(
    path: str,
    chainer: Chainer,
    group: str,
    mode: str = "retained",
    position=None,
    **kwargs,
) -> pd.DataFrame:
    """
    Retrieve an individual signal.

    Applies filtering if requested and adjusts indices.
    """
    if position is None:
        # name of h5 file
        position = chainer.stem
    if mode == "retained":
        combined = chainer.retained(path, **kwargs)
    elif mode == "raw":
        combined = chainer.get_raw(path, **kwargs)
    elif mode == "daughters":
        combined = chainer.get_raw(path, **kwargs)
        combined = combined.loc[
            combined.index.get_level_values("mother_label") > 0
        ]
    elif mode == "families":
        combined = chainer[path]
    else:
        raise Exception(f"{mode} not recognised.")
    if combined is not None:
        # adjust indices
        combined["position"] = position
        combined["group"] = group
        combined.set_index(["group", "position"], inplace=True, append=True)
        combined.index = combined.index.swaplevel(-2, 0).swaplevel(-1, 1)
    # should there be an error message if None is returned?
    return combined


class MultiGrouper:
    """Wrap results from multiple experiments stored as folders inside a
    folder."""

    def __init__(self, source: Union[str, list]):
        """
        Create NameGroupers for each experiment.

        Parameters
        ----------
        source: list of str
            List of folders, one per experiment, containing h5 files.
        """
        if isinstance(source, str):
            source = Path(source)
            self.exp_dirs = list(source.glob("*"))
        else:
            self.exp_dirs = [Path(x) for x in source]
        self.groupers = [NameGrouper(d) for d in self.exp_dirs]
        for group in self.groupers:
            group.load_chains()

    @property
    def available(self) -> None:
        """Print available signals and number of chains, one per position, for each Grouper."""
        for gpr in self.groupers:
            print(gpr.available_grouped)

    @property
    def sigtable(self) -> pd.DataFrame:
        """Generate a table showing the number of positions, or h5 files, available for each signal with one column per experiment."""

        def regex_cleanup(x):
            x = re.sub(r"extraction\/", "", x)
            x = re.sub(r"postprocessing\/", "", x)
            x = re.sub(r"\/max", "", x)
            return x

        if not hasattr(self, "_sigtable"):
            raw_mat = [
                [s.available for s in gpr.chainers.values()]
                for gpr in self.groupers
            ]
            available_grouped = [
                Counter([x for y in grp for x in y]) for grp in raw_mat
            ]
            nexps = len(available_grouped)
            sigs_idx = list(
                set([y for x in available_grouped for y in x.keys()])
            )
            sigs_idx = [regex_cleanup(x) for x in sigs_idx]
            nsigs = len(sigs_idx)
            sig_matrix = np.zeros((nsigs, nexps))
            for i, c in enumerate(available_grouped):
                for k, v in c.items():
                    sig_matrix[sigs_idx.index(regex_cleanup(k)), i] = v
            sig_matrix[sig_matrix == 0] = np.nan
            self._sigtable = pd.DataFrame(
                sig_matrix,
                index=sigs_idx,
                columns=[x.name for x in self.exp_dirs],
            )
        return self._sigtable

    def _sigtable_plot(self) -> None:
        """
        Plot number of chains for all available experiments.

        Examples
        --------
        FIXME: Add docs.
        """
        ax = sns.heatmap(self.sigtable, cmap="viridis")
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=10,
            ha="right",
            rotation_mode="anchor",
        )
        plt.show()

    def aggregate_signal(
        self,
        path: Union[str, list],
        **kwargs,
    ) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """
        Aggregate chains, one per position, from multiple Groupers, one per experiment.

        Parameters
        ----------
        path : Union[str, list]
            String or list of strings indicating the signal(s) to fetch.
        **kwargs :
            Passed to Grouper.concat_signal.

        Returns
        -------
        concatenated: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
            A multi-index dataFrame or a dictionary of multi-index dataframes, one per signal

        Examples
        --------
        >>> mg = MultiGrouper(["pHCalibrate7_24", "pHCalibrate6_7"])
        >>> p405 = mg.aggregate_signal("extraction/pHluorin405_0_4/max/median")
        >>> p588 = mg.aggregate_signal("extraction/pHluorin488_0_4/max/median")
        >>> ratio = p405 / p488
        """
        if isinstance(path, str):
            path = [path]
        sigs = {s: [] for s in path}
        for s in path:
            for grp in self.groupers:
                try:
                    sigset = grp.concat_signal(s, **kwargs)
                    new_idx = pd.MultiIndex.from_tuples(
                        [(grp.name, *x) for x in sigset.index],
                        names=("experiment", *sigset.index.names),
                    )
                    sigset.index = new_idx
                    sigs[s].append(sigset)
                except Exception as e:
                    print("Grouper {} failed: {}".format(grp.name, e))
        concatenated = {
            name: pd.concat(multiexp_sig)
            for name, multiexp_sig in sigs.items()
        }
        if len(concatenated) == 1:
            concatenated = list(concatenated.values())[0]
        return concatenated



"""
Tools to interact with h5 files and handle data consistently.
"""
import collections
import logging
import typing as t
from itertools import chain, groupby, product
from typing import Union

import h5py
import numpy as np
import yaml


class BridgeH5:
    """
    Base class to interact with h5 files.

    It includes functions that predict how long segmentation will take.
    """

    def __init__(self, filename, flag="r"):
        """Initialise with the name of the h5 file."""
        self.filename = filename
        if flag is not None:
            self._hdf = h5py.File(filename, flag)
            self._filecheck

    def _log(self, message: str, level: str = "warn"):
        # Log messages in the corresponding level
        logger = logging.getLogger("aliby")
        getattr(logger, level)(f"{self.__class__.__name__}: {message}")

    def _filecheck(self):
        assert "cell_info" in self._hdf, "Invalid file. No 'cell_info' found."

    def close(self):
        """Close the h5 file."""
        self._hdf.close()

    @property
    def meta_h5(self) -> t.Dict[str, t.Any]:
        """Return metadata, defining it if necessary."""
        if not hasattr(self, "_meta_h5"):
            with h5py.File(self.filename, "r") as f:
                self._meta_h5 = dict(f.attrs)
        return self._meta_h5

    @property
    def cell_tree(self):
        return self.get_info_tree()

    @staticmethod
    def get_consecutives(tree, nstepsback):
        """Receives a sorted tree and returns the keys of consecutive elements."""
        # get tp level
        vals = {k: np.array(list(v)) for k, v in tree.items()}
        # get indices of consecutive elements
        where_consec = [
            {
                k: np.where(np.subtract(v[n + 1 :], v[: -n - 1]) == n + 1)[0]
                for k, v in vals.items()
            }
            for n in range(nstepsback)
        ]
        return where_consec

    def get_npairs(self, nstepsback=2, tree=None):
        if tree is None:
            tree = self.cell_tree
        consecutive = self.get_consecutives(tree, nstepsback=nstepsback)
        flat_tree = flatten(tree)
        n_predictions = 0
        for i, d in enumerate(consecutive, 1):
            flat = list(chain(*[product([k], list(v)) for k, v in d.items()]))
            pairs = [(f, (f[0], f[1] + i)) for f in flat]
            for p in pairs:
                n_predictions += len(flat_tree.get(p[0], [])) * len(
                    flat_tree.get(p[1], [])
                )
        return n_predictions

    def get_npairs_over_time(self, nstepsback=2):
        tree = self.cell_tree
        npairs = []
        for tp in self._hdf["cell_info"]["processed_timepoints"][()]:
            tmp_tree = {
                k: {k2: v2 for k2, v2 in v.items() if k2 <= tp}
                for k, v in tree.items()
            }
            npairs.append(self.get_npairs(tree=tmp_tree))
        return np.diff(npairs)

    def get_info_tree(
        self, fields: Union[tuple, list] = ("trap", "timepoint", "cell_label")
    ):
        """
        Return traps, time points and labels for this position in the form of a tree in the hierarchy determined by the argument fields.

        Note that it is compressed to non-empty elements and timepoints.

        Default hierarchy is:
        - trap
        - time point
        - cell label

        This function currently produces trees of depth 3, but it can easily be extended for deeper trees if needed (e.g. considering groups, chambers and/or positions).

        Parameters
        ----------
        fields: list of strs
            Fields to fetch from 'cell_info' inside the h5 file.

        Returns
        ----------
        Nested dictionary where keys (or branches) are the upper levels and the leaves are the last element of :fields:.
        """
        zipped_info = (*zip(*[self._hdf["cell_info"][f][()] for f in fields]),)
        return recursive_groupsort(zipped_info)


def groupsort(iterable: Union[tuple, list]):
    """Sorts iterable and returns a dictionary where the values are grouped by the first element."""
    iterable = sorted(iterable, key=lambda x: x[0])
    grouped = {
        k: [x[1:] for x in v] for k, v in groupby(iterable, lambda x: x[0])
    }
    return grouped


def recursive_groupsort(iterable):
    """Recursive extension of groupsort."""
    if len(iterable[0]) > 1:
        return {
            k: recursive_groupsort(v) for k, v in groupsort(iterable).items()
        }
    else:
        # only two elements in list
        return [x[0] for x in iterable]


def flatten(d, parent_key="", sep="_"):
    """Flatten nested dict. Adapted from https://stackoverflow.com/a/6027615."""
    items = []
    for k, v in d.items():
        new_key = parent_key + (k,) if parent_key else (k,)
        if isinstance(v, collections.MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def attrs_from_h5(fpath: str):
    """Return attributes as dict from an h5 file."""
    with h5py.File(fpath, "r") as f:
        return dict(f.attrs)


def image_creds_from_h5(fpath: str):
    """Return image id and server credentials from an h5."""
    attrs = attrs_from_h5(fpath)
    return (
        attrs["image_id"],
        {
            k: yaml.safe_load(attrs["parameters"])["general"][k]
            for k in ("username", "password", "host")
        },
    )



# If you publish results that make use of this software or the Birth Annotator
# for Budding Yeast algorithm, please cite:
# Julian M J Pietsch, Alán Muñoz, Diane Adjavon, Ivan B N Clark, Peter S
# Swain, 2021, Birth Annotator for Budding Yeast (in preparation).
#
#
# The MIT License (MIT)
#
# Copyright (c) Julian Pietsch, Alán Muñoz and Diane Adjavon 2021
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
import typing as t

import numpy as np


# Calculate barycentre
def calc_barycentre(centres, weights=None, **kwargs):
    """
    :centres: ndarray containing the (x,y) centres of each cell
    :weights: (optional) list of weights to consider for each cell
    """
    if weights is None:
        weights = np.ones_like(centres)

    barycentre = np.average(centres, axis=0, weights=weights)
    return barycentre


# Calculate distance to center
def calc_barydists(centres, bary, **kwargs):
    """
    Calculate distances to the barycentre
    :centre: int (2,) tuple. Centre of cell
    :bary: float (2,) tuple. Barycentre of image
    """
    vec2bary = centres - bary
    dists = np.sqrt(np.sum(vec2bary**2, axis=1))

    return dists


# Calculate angle to center
def calc_baryangles(centres, bary, areas=None, **kwargs):
    """
    Calculate angle using centre of cell and barycentre
    :centre: int (2,) tuple. Centre of cell
    :bary: float (2,) tuple. Barycentre of image
    :anchor_cell: int Cell id to use as angle 0.
    """

    angles = []
    vec2bary = centres - bary
    angles = np.apply_along_axis(lambda x: np.arctan2(*x), 1, vec2bary)
    if areas is not None:
        anchor_cell = np.argmax(areas)
        angles -= angles[anchor_cell]

    return angles


def pick_baryfun(key):
    baryfuns = {"barydist": calc_barydists, "baryangle": calc_baryangles}
    return baryfuns[key]


##  Tracking benchmark utils


def lol_to_adj(cell_ids: t.List[t.List[int]]):
    """
    Convert a series list of lists with cell ids into a matrix
    representing a graph.

    Note that information is lost in the process, and a matrix can't be
    turned back into a list of list by itself.

    input

    :lol: list of lists with cell ids

    returns

    :adj_matrix: (n, n) ndarray where n is the number of cells

    """
    n = len([y for x in cell_ids for y in x])
    adj_mat = np.zeros((n, n))

    prev = None
    cur = 0
    for c_ids_single_lst in cell_ids:
        if not prev:
            prev = c_ids_single_lst
        else:
            for i, el in enumerate(c_ids_single_lst):
                prev_idx = prev.index(el) if el in prev else None
                if prev_idx is not None:
                    adj_mat[cur + len(prev) + i, cur + prev_idx] = True
            cur += len(c_ids_single_lst)

    return adj_mat


def compare_pred_truth_lols(prediction, truth):
    """
    input

    :prediction: list of lists with predicted cell ids
    :truth: list of lists with real cell ids

    returns

    number of diferences between equivalent truth matrices

    """
    adj_pred = lol_to_adj(prediction)
    adj_truth = lol_to_adj(truth)

    return int(((adj_pred - adj_truth) != 0).sum())



#!/usr/bin/env python3
"""
Image: Loads images and registers them.

Image instances loads images from a specified directory into an object that
also contains image properties such as name and metadata.  Pixels from images
are stored in dask arrays; the standard way is to store them in 5-dimensional
arrays: T(ime point), C(channel), Z(-stack), Y, X.

This module consists of a base Image class (BaseLocalImage).  ImageLocalOME
handles local OMERO images.  ImageDir handles cases in which images are split
into directories, with each time point and channel having its own image file.
ImageDummy is a dummy class for silent failure testing.
"""

import typing as t
from abc import ABC, abstractmethod, abstractproperty
from datetime import datetime
from pathlib import Path

import dask.array as da
import numpy as np
import xmltodict
import zarr
from dask.array.image import imread
from importlib_resources import files
from tifffile import TiffFile

from agora.io.metadata import dir_to_meta, dispatch_metadata_parser


def get_examples_dir():
    """Get examples directory which stores dummy image for tiler"""
    return files("aliby").parent.parent / "examples" / "tiler"


def instantiate_image(
    source: t.Union[str, int, t.Dict[str, str], Path], **kwargs
):
    """Wrapper to instatiate the appropiate image

    Parameters
    ----------
    source : t.Union[str, int, t.Dict[str, str], Path]
        Image identifier

    Examples
    --------
    image_path = "path/to/image"]
    with instantiate_image(image_path) as img:
        print(imz.data, img.metadata)

    """
    return dispatch_image(source)(source, **kwargs)


def dispatch_image(source: t.Union[str, int, t.Dict[str, str], Path]):
    """
    Wrapper to pick the appropiate Image class depending on the source of data.
    """
    if isinstance(source, (int, np.int64)):
        from aliby.io.omero import Image

        instatiator = Image
    elif isinstance(source, dict) or (
        isinstance(source, (str, Path)) and Path(source).is_dir()
    ):
        if Path(source).suffix == ".zarr":
            instatiator = ImageZarr
        else:
            instatiator = ImageDir
    elif isinstance(source, str) and Path(source).is_file():
        instatiator = ImageLocalOME
    else:
        raise Exception(f"Invalid data source at {source}")

    return instatiator


class BaseLocalImage(ABC):
    """
    Base Image class to set path and provide context management method.
    """

    _default_dimorder = "tczyx"

    def __init__(self, path: t.Union[str, Path]):
        # If directory, assume contents are naturally sorted
        self.path = Path(path)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        for e in exc:
            if e is not None:
                print(e)
        return False

    def rechunk_data(self, img):
        # Format image using x and y size from metadata.

        self._rechunked_img = da.rechunk(
            img,
            chunks=(
                1,
                1,
                1,
                self._meta["size_y"],
                self._meta["size_x"],
            ),
        )
        return self._rechunked_img

    @abstractmethod
    def get_data_lazy(self) -> da.Array:
        pass

    @abstractproperty
    def name(self):
        pass

    @abstractproperty
    def dimorder(self):
        pass

    @property
    def data(self):
        return self.get_data_lazy()

    @property
    def metadata(self):
        return self._meta

    def set_meta(self):
        """Load metadata using parser dispatch"""
        self._meta = dispatch_metadata_parser(self.path)


class ImageLocalOME(BaseLocalImage):
    """
    Local OMERO Image class.

    This is a derivative Image class. It fetches an image from OMEXML data format,
    in which a multidimensional tiff image contains the metadata.
    """

    def __init__(self, path: str, dimorder=None):
        super().__init__(path)
        self._id = str(path)

    def set_meta(self):
        meta = dict()
        try:
            with TiffFile(path) as f:
                self._meta = xmltodict.parse(f.ome_metadata)["OME"]

            for dim in self.dimorder:
                meta["size_" + dim.lower()] = int(
                    self._meta["Image"]["Pixels"]["@Size" + dim]
                )
                meta["channels"] = [
                    x["@Name"]
                    for x in self._meta["Image"]["Pixels"]["Channel"]
                ]
                meta["name"] = self._meta["Image"]["@Name"]
                meta["type"] = self._meta["Image"]["Pixels"]["@Type"]

        except Exception as e:  # Images not in OMEXML

            print("Warning:Metadata not found: {}".format(e))
            print(
                f"Warning: No dimensional info provided. Assuming {self._default_dimorder}"
            )

            # Mark non-existent dimensions for padding
            self.base = self._default_dimorder
            # self.ids = [self.index(i) for i in dimorder]

            self._dimorder = base

            self._meta = meta

    @property
    def name(self):
        return self._meta["name"]

    @property
    def date(self):
        date_str = [
            x
            for x in self._meta["StructuredAnnotations"]["TagAnnotation"]
            if x["Description"] == "Date"
        ][0]["Value"]
        return datetime.strptime(date_str, "%d-%b-%Y")

    @property
    def dimorder(self):
        """Order of dimensions in image"""
        if not hasattr(self, "_dimorder"):
            self._dimorder = self._meta["Image"]["Pixels"]["@DimensionOrder"]
        return self._dimorder

    @dimorder.setter
    def dimorder(self, order: str):
        self._dimorder = order
        return self._dimorder

    def get_data_lazy(self) -> da.Array:
        """Return 5D dask array. For lazy-loading  multidimensional tiff files"""

        if not hasattr(self, "formatted_img"):
            if not hasattr(self, "ids"):  # Standard dimension order
                img = (imread(str(self.path))[0],)
            else:  # Custom dimension order, we rearrange the axes for compatibility
                img = imread(str(self.path))[0]
                for i, d in enumerate(self._dimorder):
                    self._meta["size_" + d.lower()] = img.shape[i]

                target_order = (
                    *self.ids,
                    *[
                        i
                        for i, d in enumerate(self.base)
                        if d not in self.dimorder
                    ],
                )
                reshaped = da.reshape(
                    img,
                    shape=(
                        *img.shape,
                        *[1 for _ in range(5 - len(self.dimorder))],
                    ),
                )
                img = da.moveaxis(
                    reshaped, range(len(reshaped.shape)), target_order
                )

        return self.rechunk_data(img)


class ImageDir(BaseLocalImage):
    """
    Image class for the case in which all images are split in one or
    multiple folders with time-points and channels as independent files.
    It inherits from BaseLocalImage so we only override methods that are critical.

    Assumptions:
    - One folders per position.
    - Images are flat.
    - Channel, Time, z-stack and the others are determined by filenames.
    - Provides Dimorder as it is set in the filenames, or expects order during instatiation
    """

    def __init__(self, path: t.Union[str, Path], **kwargs):
        super().__init__(path)
        self.image_id = str(self.path.stem)

        self._meta = dir_to_meta(self.path)

    def get_data_lazy(self) -> da.Array:
        """Return 5D dask array. For lazy-loading local multidimensional tiff files"""

        img = imread(str(self.path / "*.tiff"))

        # If extra channels, pick the first stack of the last dimensions

        while len(img.shape) > 3:
            img = img[..., 0]

        if self._meta:
            self._meta["size_x"], self._meta["size_y"] = img.shape[-2:]

            # Reshape using metadata
            # img = da.reshape(img, (*self._meta, *img.shape[1:]))
            img = da.reshape(img, self._meta.values())
            original_order = [
                i[-1] for i in self._meta.keys() if i.startswith("size")
            ]
            # Swap axis to conform with normal order
            target_order = [
                self._default_dimorder.index(x) for x in original_order
            ]
            img = da.moveaxis(
                img,
                list(range(len(original_order))),
                target_order,
            )
            pixels = self.rechunk_data(img)
        return pixels

    @property
    def name(self):
        return self.path.stem

    @property
    def dimorder(self):
        # Assumes only dimensions start with "size"
        return [
            k.split("_")[-1] for k in self._meta.keys() if k.startswith("size")
        ]


class ImageZarr(BaseLocalImage):
    """
    Read zarr compressed files.
    These are outputed by the script
    skeletons/scripts/howto_omero/convert_clone_zarr_to_tiff.py
    """

    def __init__(self, path: t.Union[str, Path], **kwargs):
        super().__init__(path)
        self.set_meta()
        try:
            self._img = zarr.open(self.path)
            self.add_size_to_meta()
        except Exception as e:
            print(f"Could not add size info to metadata: {e}")

    def get_data_lazy(self) -> da.Array:
        """Return 5D dask array. For lazy-loading local multidimensional zarr files"""
        return self._img

    def add_size_to_meta(self):
        self._meta.update(
            {
                f"size_{dim}": shape
                for dim, shape in zip(self.dimorder, self._img.shape)
            }
        )

    @property
    def name(self):
        return self.path.stem

    @property
    def dimorder(self):
        # FIXME hardcoded order based on zarr compression/cloning script
        return "TCZYX"
        # Assumes only dimensions start with "size"
        # return [
        #     k.split("_")[-1] for k in self._meta.keys() if k.startswith("size")
        # ]


class ImageDummy(BaseLocalImage):
    """
    Dummy Image class.

    ImageDummy mimics the other Image classes in such a way that it is accepted
    by Tiler.  The purpose of this class is for testing, in particular,
    identifying silent failures.  If something goes wrong, we should be able to
    know whether it is because of bad parameters or bad input data.

    For the purposes of testing parameters, ImageDummy assumes that we already
    know the tiler parameters before Image instances are instantiated.  This is
    true for a typical pipeline run.
    """

    def __init__(self, tiler_parameters: dict):
        """Builds image instance

        Parameters
        ----------
        tiler_parameters : dict
            Tiler parameters, in dict form. Following
            aliby.tile.tiler.TilerParameters, the keys are: "tile_size" (size of
            tile), "ref_channel" (reference channel for tiling), and "ref_z"
            (reference z-stack, 0 to choose a default).
        """
        self.ref_channel = tiler_parameters["ref_channel"]
        self.ref_z = tiler_parameters["ref_z"]

    # Goal: make Tiler happy.
    @staticmethod
    def pad_array(
        image_array: da.Array,
        dim: int,
        n_empty_slices: int,
        image_position: int = 0,
    ):
        """Extends a dimension in a dask array and pads with zeros

        Extends a dimension in a dask array that has existing content, then pads
        with zeros.

        Parameters
        ----------
        image_array : da.Array
            Input dask array
        dim : int
            Dimension in which to extend the dask array.
        n_empty_slices : int
            Number of empty slices to extend the dask array by, in the specified
            dimension/axis.
        image_position : int
            Position within the new dimension to place the input arary, default 0
            (the beginning).

        Examples
        --------
        ```
        extended_array = pad_array(
            my_da_array, dim = 2, n_empty_slices = 4, image_position = 1)
        ```
        Extends a dask array called `my_da_array` in the 3rd dimension
        (dimensions start from 0) by 4 slices, filled with zeros.  And puts the
        original content in slice 1 of the 3rd dimension
        """
        # Concats zero arrays with same dimensions as image_array, and puts
        # image_array as first element in list of arrays to be concatenated
        zeros_array = da.zeros_like(image_array)
        return da.concatenate(
            [
                *([zeros_array] * image_position),
                image_array,
                *([zeros_array] * (n_empty_slices - image_position)),
            ],
            axis=dim,
        )

    # Logic: We want to return a image instance
    def get_data_lazy(self) -> da.Array:
        """Return 5D dask array. For lazy-loading multidimensional tiff files. Dummy image."""
        examples_dir = get_examples_dir()
        # TODO: Make this robust to having multiple TIFF images, one for each z-section,
        # all falling under the same "pypipeline_unit_test_00_000001_Brightfield_*.tif"
        # naming scheme.  The aim is to create a multidimensional dask array that stores
        # the z-stacks.
        img_filename = "pypipeline_unit_test_00_000001_Brightfield_003.tif"
        img_path = examples_dir / img_filename
        # img is a dask array has three dimensions: z, x, y
        # TODO: Write a test to confirm this: If everything worked well,
        # z = 1, x = 1200, y = 1200
        img = imread(str(img_path))
        # Adds t & c dimensions
        img = da.reshape(
            img, (1, 1, img.shape[-3], img.shape[-2], img.shape[-1])
        )
        # Pads t, c, and z dimensions
        img = self.pad_array(
            img, dim=0, n_empty_slices=199
        )  # 200 timepoints total
        img = self.pad_array(img, dim=1, n_empty_slices=2)  # 3 channels
        img = self.pad_array(
            img, dim=2, n_empty_slices=4, image_position=self.ref_z
        )  # 5 z-stacks
        return img

    @property
    def name(self):
        pass

    @property
    def dimorder(self):
        pass



#!/usr/bin/env jupyter
"""
Utilities based on association are used to efficiently acquire indices of tracklets with some kind of relationship.
This can be:
    - Cells that are to be merged
    - Cells that have a linear relationship
"""

import numpy as np
import typing as t


def validate_association(
    association: np.ndarray,
    indices: np.ndarray,
    match_column: t.Optional[int] = None,
) -> t.Tuple[np.ndarray, np.ndarray]:

    """Select rows from the first array that are present in both.
        We use casting for fast multiindexing, generalising for lineage dynamics


        Parameters
        ----------
        association : np.ndarray
            2-D array where columns are (trap, mother, daughter) or 3-D array where
            dimensions are (X,trap,2), containing tuples ((trap,mother), (trap,daughter))
            across the 3rd dimension.
        indices : np.ndarray
            2-D array where each column is a different level. This should not include mother_label.
        match_column: int
            int indicating a specific column is required to match (i.e.
            0-1 for target-source when trying to merge tracklets or mother-bud for lineage)
            must be present in indices. If it is false one match suffices for the resultant indices
            vector to be True.

        Returns
        -------
        np.ndarray
            1-D boolean array indicating valid merge events.
        np.ndarray
            1-D boolean array indicating indices with an association relationship.

        Examples
        --------

        >>> import numpy as np
        >>> from agora.utils.indexing import validate_association
        >>> merges = np.array(range(12)).reshape(3,2,2)
        >>> indices = np.array(range(6)).reshape(3,2)

        >>> print(merges, indices)
        >>> print(merges); print(indices)
        [[[ 0  1]
          [ 2  3]]

         [[ 4  5]
          [ 6  7]]

         [[ 8  9]
          [10 11]]]

        [[0 1]
         [2 3]
         [4 5]]

        >>> valid_associations, valid_indices  = validate_association(merges, indices)
        >>> print(valid_associations, valid_indices)
    [ True False False] [ True  True False]

    """
    if association.ndim == 2:
        # Reshape into 3-D array for broadcasting if neded
        # association = np.stack(
        #     (association[:, [0, 1]], association[:, [0, 2]]), axis=1
        # )
        association = _assoc_indices_to_3d(association)

    # Compare existing association with available indices
    # Swap trap and label axes for the association array to correctly cast
    valid_ndassociation = association[..., None] == indices.T[None, ...]

    # Broadcasting is confusing (but efficient):
    # First we check the dimension across trap and cell id, to ensure both match
    valid_cell_ids = valid_ndassociation.all(axis=2)

    if match_column is None:
        # Then we check the merge tuples to check which cases have both target and source
        valid_association = valid_cell_ids.any(axis=2).all(axis=1)

        # Finally we check the dimension that crosses all indices, to ensure the pair
        # is present in a valid merge event.
        valid_indices = (
            valid_ndassociation[valid_association].all(axis=2).any(axis=(0, 1))
        )
    else:  # We fetch specific indices if we aim for the ones with one present
        valid_indices = valid_cell_ids[:, match_column].any(axis=0)
        # Valid association then becomes a boolean array, true means that there is a
        # match (match_column) between that cell and the index
        valid_association = (
            valid_cell_ids[:, match_column] & valid_indices
        ).any(axis=1)

    return valid_association, valid_indices


def _assoc_indices_to_3d(ndarray: np.ndarray):
    """
    Convert the last column to a new row while repeating all previous indices.

    This is useful when converting a signal multiindex before comparing association.

    Assumes the input array has shape (N,3)
    """
    result = ndarray
    if len(ndarray) and ndarray.ndim > 1:
        if ndarray.shape[1] == 3:  # Faster indexing for single positions
            result = np.transpose(
                np.hstack((ndarray[:, [0]], ndarray)).reshape(-1, 2, 2),
                axes=[0, 2, 1],
            )
        else:  # 20% slower but more general indexing
            columns = np.arange(ndarray.shape[1])

            result = np.stack(
                (
                    ndarray[:, np.delete(columns, -1)],
                    ndarray[:, np.delete(columns, -2)],
                ),
                axis=1,
            )
    return result


def _3d_index_to_2d(array: np.ndarray):
    """
    Opposite to _assoc_indices_to_3d.
    """
    result = array
    if len(array):
        result = np.concatenate(
            (array[:, 0, :], array[:, 1, 1, np.newaxis]), axis=1
        )
    return result


def compare_indices(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Fetch two 2-D indices and return a binary 2-D matrix
    where a True value links two cells where all cells are the same
    """
    return (x[..., None] == y.T[None, ...]).all(axis=1)



#!/usr/bin/env python3

import numpy as np
from matplotlib import cm, colors


def generate_palette_map(df):
    """Create a palette map based on the strains in a dataframe"""
    strain_list = np.unique(df.index.get_level_values("strain"))
    palette_cm = cm.get_cmap("Set1", len(strain_list) + 1)
    palette_rgb = [
        colors.rgb2hex(palette_cm(index / len(strain_list))[:3])
        for index, _ in enumerate(strain_list)
    ]
    palette_map = dict(zip(strain_list, palette_rgb))
    return palette_map



#!/usr/bin/env jupyter

import re
import typing as t
from copy import copy

import pandas as pd

from agora.io.signal import Signal
from agora.utils.kymograph import bidirectional_retainment_filter
from postprocessor.core.abc import get_process


class Chainer(Signal):
    """
    Extend Signal by applying post-processes and allowing composite signals that combine basic signals.
    It "chains" multiple processes upon fetching a dataset to produce the desired datasets.

    Instead of reading processes previously applied, it executes
    them when called.
    """

    _synonyms = {
        "m5m": ("extraction/GFP/max/max5px", "extraction/GFP/max/median")
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def replace_path(path: str, bgsub: bool = ""):
            # function to add bgsub to paths
            channel = path.split("/")[1]
            suffix = "_bgsub" if bgsub else ""
            path = re.sub(channel, f"{channel}{suffix}", path)
            return path

        # Add chain with and without bgsub for composite statistics
        self.common_chains = {
            alias
            + bgsub: lambda **kwargs: self.get(
                replace_path(denominator, alias + bgsub), **kwargs
            )
            / self.get(replace_path(numerator, alias + bgsub), **kwargs)
            for alias, (denominator, numerator) in self._synonyms.items()
            for bgsub in ("", "_bgsub")
        }

    def get(
        self,
        dataset: str,
        chain: t.Collection[str] = ("standard", "interpolate", "savgol"),
        in_minutes: bool = True,
        stages: bool = True,
        retain: t.Optional[float] = None,
        **kwargs,
    ):
        """Load data from an h5 file."""
        if dataset in self.common_chains:
            # get dataset for composite chains
            data = self.common_chains[dataset](**kwargs)
        else:
            # use Signal's get_raw
            data = self.get_raw(dataset, in_minutes=in_minutes, lineage=True)
            if chain:
                data = self.apply_chain(data, chain, **kwargs)
        if retain:
            # keep data only from early time points
            data = self.get_retained(data, retain)
        if stages and "stage" not in data.columns.names:
            # return stages as additional column level
            stages_index = [
                x
                for i, (name, span) in enumerate(self.stages_span_tp)
                for x in (f"{i} { name }",) * span
            ]
            data.columns = pd.MultiIndex.from_tuples(
                zip(stages_index, data.columns),
                names=("stage", "time"),
            )
        return data

    def apply_chain(
        self, input_data: pd.DataFrame, chain: t.Tuple[str, ...], **kwargs
    ):
        """
        Apply a series of processes to a data set.

        Like postprocessing, Chainer consecutively applies processes.

        Parameters can be passed as kwargs.

        Chainer does not support applying the same process multiple times with different parameters.

        Parameters
        ----------
        input_data : pd.DataFrame
            Input data to process.
        chain : t.Tuple[str, ...]
            Tuple of strings with the names of the processes
        **kwargs : kwargs
            Arguments passed on to Process.as_function() method to modify the parameters.

        Examples
        --------
        FIXME: Add docs.


        """
        result = copy(input_data)
        self._intermediate_steps = []
        for process in chain:
            if process == "standard":
                result = bidirectional_retainment_filter(result)
            else:
                params = kwargs.get(process, {})
                process_cls = get_process(process)
                result = process_cls.as_function(result, **params)
                process_type = process_cls.__module__.split(".")[-2]
                if process_type == "reshapers":
                    if process == "merger":
                        raise (NotImplementedError)
            self._intermediate_steps.append(result)
        return result



#!/usr/bin/env jupyter
"""
Basic plotting functions for cell visualisation
"""

import typing as t

import numpy as np
from grid_strategy import strategies
from matplotlib import pyplot as plt


def plot_overlay(
    bg: np.ndarray, fg: np.ndarray, alpha: float = 1.0, ax=plt
) -> None:
    """
    Plot two images, one on top of the other.
    """

    ax1 = ax.imshow(
        bg, cmap=plt.cm.gray, interpolation="none", interpolation_stage="rgba"
    )
    ax2 = ax.imshow(
        stretch(fg),
        alpha=alpha,
        interpolation="none",
        interpolation_stage="rgba",
    )
    plt.axis("off")
    return ax1, ax2


def plot_overlay_in_square(data: t.Tuple[np.ndarray, np.ndarray]):
    """
    Plot images in an automatically-arranged grid.
    """
    specs = strategies.SquareStrategy("center").get_grid(len(data))
    for i, (gs, (tile, mask)) in enumerate(zip(specs, data)):
        ax = plt.subplot(gs)
        plot_overlay(tile, mask, ax=ax)


def plot_in_square(data: t.Iterable):
    """
    Plot images in an automatically-arranged grid. Only takes one mask
    """
    specs = strategies.SquareStrategy("center").get_grid(len(data))
    for i, (gs, datum) in enumerate(zip(specs, data)):
        ax = plt.subplot(gs)
        ax.imshow(datum)


def stretch_clip(image, clip=True):
    """
    Performs contrast stretching on an input image.

    This function takes an array-like input image and enhances its contrast by adjusting
    the dynamic range of pixel values. It first scales the pixel values between 0 and 255,
    then clips the values that are below the 2nd percentile or above the 98th percentile.
    Finally, the pixel values are scaled to the range between 0 and 1.

    Parameters
    ----------
    image : array-like
        Input image.

    Returns
    -------
    stretched : ndarray
        Contrast-stretched version of the input image.

    Examples
    --------
    FIXME: Add docs.
    """
    from copy import copy

    image = image[~np.isnan(image)]
    image = ((image - image.min()) / (image.max() - image.min())) * 255
    if clip:
        minval = np.percentile(image, 2)
        maxval = np.percentile(image, 98)
        image = np.clip(image, minval, maxval)
    image = (image - minval) / (maxval - minval)
    return image


def stretch(image):

    nona = image[~np.isnan(image)]

    return (image - nona.min()) / (nona.max() - nona.min())



#!/usr/bin/env python

import pandas as pd
from sklearn.preprocessing import StandardScaler

from agora.abc import ParametersABC
from postprocessor.core.abc import PostProcessABC


class standardscalerParameters(ParametersABC):
    """
    Parameters for the 'scale' process.
    """

    _defaults = {}


class standardscaler(PostProcessABC):
    """
    Process to scale a DataFrame of a signal using the standard scaler.

    Methods
    -------
    run(signal: pd.DataFrame)
        Scale values in a dataframe of time series.
    """

    def __init__(self, parameters: standardscalerParameters):
        super().__init__(parameters)

    def run(self, signal: pd.DataFrame):
        """Scale values in a dataframe of time series.

        Scale values in a dataframe of time series.  This function is effectively a
        wrapper for sklearn.preprocessing.StandardScaler.

        Parameters
        ----------
        signal : pd.DataFrame
            Time series, with rows indicating individual time series (e.g. from
            each cell), and columns indicating time points.
        """
        signal_array = signal.to_numpy()
        scaler = StandardScaler().fit(signal_array.transpose())
        signal_scaled_array = scaler.transform(signal_array.transpose())
        signal_scaled_array = signal_scaled_array.transpose()
        signal_scaled = pd.DataFrame(
            signal_scaled_array, columns=signal.columns, index=signal.index
        )
        return signal_scaled



import typing as t
from typing import Dict, Tuple

import numpy as np
import pandas as pd

from agora.utils.lineage import mb_array_to_dict
from postprocessor.core.lineageprocess import (
    LineageProcess,
    LineageProcessParameters,
)


class BudMetricParameters(LineageProcessParameters):
    """
    Parameters
    """

    _defaults = {"lineage_location": "postprocessing/lineage_merged"}


class BudMetric(LineageProcess):
    """
    Requires mother-bud information to create a new dataframe where the indices are mother ids and
    values are the daughters' values for a given signal.
    """

    def __init__(self, parameters: BudMetricParameters):
        super().__init__(parameters)

    def run(
        self,
        signal: pd.DataFrame,
        lineage: Dict[pd.Index, Tuple[pd.Index]] = None,
    ):
        if lineage is None:
            if hasattr(self, "lineage"):
                lineage = self.lineage
            else:
                assert "mother_label" in signal.index.names
                lineage = signal.index.to_list()

        return self.get_bud_metric(signal, mb_array_to_dict(lineage))

    @staticmethod
    def get_bud_metric(
        signal: pd.DataFrame, md: Dict[Tuple, Tuple[Tuple]] = None
    ):
        """

        signal: Daughter-inclusive dataframe
        md: Mother-daughters dictionary where key is mother's index and value a list of daugher indices

        Get fvi (First Valid Index) for all cells
        Create empty matrix
        for every mother:
         - Get daughters' subdataframe
         - sort  daughters by cell label
         - get series of fvis
         - concatenate the values of these ranges from the dataframe
        Fill the empty matrix
        Convert matrix into dataframe using mother indices

        """
        mothers_mat = np.zeros((len(md), signal.shape[1]))
        cells_were_dropped = 0  # Flag determines if mothers (1), daughters (2) or both were missing (3)

        md_index = signal.index
        if (
            "mother_label" not in md_index.names
        ):  # Generate mother label from md dict if unavailable
            d = {v: k for k, values in md.items() for v in values}
            signal["mother_label"] = list(
                map(lambda x: d.get(x, [0])[-1], signal.index)
            )
            signal.set_index("mother_label", append=True, inplace=True)
            related_items = set(
                [*md.keys(), *[y for x in md.values() for y in x]]
            )
            md_index = md_index.intersection(related_items)
        elif "mother_label" in md_index.names:
            md_index = md_index.droplevel("mother_label")
        else:
            raise ("Unavailable relationship information")

        if len(md_index) < len(signal):
            print("Dropped cells before bud_metric")  # TODO log

        signal = (
            signal.reset_index("mother_label")
            .loc(axis=0)[md_index]
            .set_index("mother_label", append=True)
        )

        names = list(signal.index.names)
        del names[-2]

        output_df = (
            signal.loc[signal.index.get_level_values("mother_label") > 0]
            .groupby(names)
            .apply(lambda x: _combine_daughter_tracks(x))
        )
        output_df.columns = signal.columns
        output_df["padding_level"] = 0
        output_df.set_index("padding_level", append=True, inplace=True)

        if len(output_df):
            output_df.index.names = signal.index.names
        return output_df


def _combine_daughter_tracks(tracks: t.Collection[pd.Series]):
    """
    Combine multiple time series of cells into one, overwriting values
    prioritising the most recent entity.
    """
    sorted_da_ids = tracks.sort_index(level="cell_label")
    sorted_da_ids.index = range(len(sorted_da_ids))
    tp_fvt = sorted_da_ids.apply(lambda x: x.first_valid_index(), axis=0)
    tp_fvt = sorted_da_ids.columns.get_indexer(tp_fvt)
    tp_fvt[tp_fvt < 0] = len(sorted_da_ids) - 1

    _metric = np.choose(tp_fvt, sorted_da_ids.values)
    return pd.Series(_metric, index=tracks.columns)



"""
Extracted from the baby repository. Bud Tracker algorithm to link
cell outlines as mothers and buds.
"""
# /usr/bin/env jupyter

import pickle
from os.path import join

import numpy as np
from skimage.draw import polygon

from agora.track_abc import FeatureCalculator

models_path = join(dirname(__file__), "./models")


class BudTracker(FeatureCalculator):
    def __init__(self, model=None, feats2use=None, **kwargs):

        if model is None:
            model_file = join(models_path, "mb_model_20201022.pkl")
            with open(model_file, "rb") as file_to_load:
                model = pickle.load(file_to_load)
        self.model = model

        if feats2use is None:
            feats2use = ["centroid", "area", "minor_axis_length"]
        super().__init__(feats2use, **kwargs)

        self.a_ind = self.outfeats.index("area")
        self.ma_ind = self.outfeats.index("minor_axis_length")

    ### Assign mother-
    def calc_mother_bud_stats(self, p_budneck, p_bud, masks, feats=None):
        """
        ---

        input

        :p_budneck: 2d ndarray (size_x, size_y) giving the probability that a
            pixel corresponds to a bud neck
        :p_bud: 2d ndarray (size_x, size_y) giving the probability that a pixel
            corresponds to a bud
        :masks: 3d ndarray (ncells, size_x, size_y)
        :feats: ndarray (ncells, nfeats)

        NB: ASSUMES FEATS HAVE ALREADY BEEN NORMALISED!

        returns

        :n2darray: 2d ndarray (ncells x ncells, n_feats) specifying,
            for each pair of cells in the masks array, the features used for
            mother-bud pair prediction (as per 'feats2use')
        """

        if feats is None:
            feats = self.calc_feats_from_mask(masks)
        elif len(feats) != len(masks):
            raise Exception("number of features must match number of masks")

        ncells = len(masks)

        # Entries will be NaN unless validly specified below
        p_bud_mat = np.nan * np.ones((ncells, ncells))
        p_budneck_mat = np.nan * np.ones((ncells, ncells))
        budneck_ratio_mat = np.nan * np.ones((ncells, ncells))
        size_ratio_mat = np.nan * np.ones((ncells, ncells))
        adjacency_mat = np.nan * np.ones((ncells, ncells))

        for m in range(ncells):
            for d in range(ncells):
                if m == d:
                    # Mother-bud pairs can only be between different cells
                    continue

                p_bud_mat[m, d] = np.mean(p_bud[masks[d].astype("bool")])

                a_i = self.a_ind
                size_ratio_mat[m, d] = feats[m, a_i] / feats[d, a_i]

                # Draw rectangle
                r_points = self.get_rpoints(feats, d, m)
                if r_points is None:
                    continue
                rr, cc = polygon(
                    r_points[0, :], r_points[1, :], p_budneck.shape
                )
                if len(rr) == 0:
                    # Rectangles with zero size are not informative
                    continue

                r_im = np.zeros(p_budneck.shape, dtype="bool")
                r_im[rr, cc] = True

                # Calculate the mean of bud neck probabilities greater than some threshold
                pbn = p_budneck[r_im].flatten()

                pbn = pbn[pbn > 0.2]
                p_budneck_mat[m, d] = np.mean(pbn) if len(pbn) > 0 else 0

                # Normalise number of bud-neck positive pixels by the scale of
                # the bud (a value proportional to circumference):
                raw_circumf_est = np.sqrt(feats[d, a_i]) / self.pixel_size
                budneck_ratio_mat[m, d] = pbn.sum() / raw_circumf_est

                # Adjacency is the proportion of the joining rectangle that overlaps the mother daughter union
                md_union = masks[m] | masks[d]
                adjacency_mat[m, d] = np.sum(md_union & r_im) / np.sum(r_im)

        return np.hstack(
            [
                s.flatten()[:, np.newaxis]
                for s in (
                    p_bud_mat,
                    size_ratio_mat,
                    p_budneck_mat,
                    budneck_ratio_mat,
                    adjacency_mat,
                )
            ]
        )

    def predict_mother_bud(self, p_budneck, p_bud, masks, feats=None):
        """
        ---

        input

        :p_budneck: 2d ndarray (size_x, size_y) giving the probability that a
            pixel corresponds to a bud neck
        :p_bud: 2d ndarray (size_x, size_y) giving the probability that a pixel
            corresponds to a bud
        :masks: 3d ndarray (ncells, size_x, size_y)
        :feats: ndarray (ncells, nfeats)

        returns

        :n2darray: 2d ndarray (ncells, ncells) giving probability that a cell
            (row) is a mother to another cell (column)
        """

        ncells = len(masks)

        mb_stats = self.calc_mother_bud_stats(p_budneck, p_bud, masks, feats)

        good_stats = ~np.isnan(mb_stats).any(axis=1)
        # Assume probability of bud assignment for any rows that are NaN will
        # be zero
        ba_probs = np.zeros(ncells**2)
        if good_stats.any():
            ba_probs[good_stats] = self.model.predict_proba(
                mb_stats[good_stats, :]
            )[:, 1]
        ba_probs = ba_probs.reshape((ncells,) * 2)

        return ba_probs

    def get_rpoints(self, feats, d, m):
        """
        Draw a rectangle in the budneck of cells
        ---

        NB: ASSUMES FEATS HAVE ALREADY BEEN NORMALISED!

        input

        feats: 2d ndarray (ncells, nfeats)

        returns

        r_points: 2d ndarray (2,4) with the coordinates of the rectangle corner

        """

        # Get un-scaled features for m-d pair
        descaled_feats = feats / self.pixel_size
        m_centre = descaled_feats[m, :2]
        d_centre = descaled_feats[d, :2]
        r_width = np.max((2, descaled_feats[d, self.ma_ind] * 0.25))

        # Draw connecting rectangle
        r_hvec = d_centre - m_centre
        r_wvec = np.matmul(np.array([[0, -1], [1, 0]]), r_hvec)
        r_wvec_len = np.linalg.norm(r_wvec)
        if r_wvec_len == 0:
            return None
        r_wvec = r_width * r_wvec / r_wvec_len
        r_points = np.zeros((2, 4))
        r_points[:, 0] = m_centre - 0.5 * r_wvec
        r_points[:, 1] = r_points[:, 0] + r_hvec
        r_points[:, 2] = r_points[:, 1] + r_wvec
        r_points[:, 3] = r_points[:, 2] - r_hvec

        return r_points



import typing as t

import bottleneck as bn
import h5py
import numpy as np
import pandas as pd

from agora.abc import ParametersABC, StepABC
from agora.io.cells import Cells
from agora.io.writer import Writer, load_attributes
from aliby.tile.tiler import Tiler
from extraction.core.functions.defaults import exparams_from_meta
from extraction.core.functions.distributors import reduce_z, trap_apply
from extraction.core.functions.loaders import (
    load_custom_args,
    load_funs,
    load_redfuns,
)

# define types
reduction_method = t.Union[t.Callable, str, None]
extraction_tree = t.Dict[
    str, t.Dict[reduction_method, t.Dict[str, t.Collection]]
]
extraction_result = t.Dict[
    str, t.Dict[reduction_method, t.Dict[str, t.Dict[str, pd.Series]]]
]

# Global variables used to load functions that either analyse cells or their background. These global variables both allow the functions to be stored in a dictionary for access only on demand and to be defined simply in extraction/core/functions.
CELL_FUNS, TRAPFUNS, FUNS = load_funs()
CUSTOM_FUNS, CUSTOM_ARGS = load_custom_args()
RED_FUNS = load_redfuns()

# Assign datatype depending on the metric used
# m2type = {"mean": np.float32, "median": np.ubyte, "imBackground": np.ubyte}


class ExtractorParameters(ParametersABC):
    """Base class to define parameters for extraction."""

    def __init__(
        self,
        tree: extraction_tree,
        sub_bg: set = set(),
        multichannel_ops: t.Dict = {},
    ):
        """
        Initialise.

        Parameters
        ----------
        tree: dict
            Nested dictionary indicating channels, reduction functions and
            metrics to be used.
            str channel -> U(function, None) reduction -> str metric
            If not of depth three, tree will be filled with None.
        sub_bg: set
        multichannel_ops: dict
        """
        self.tree = tree
        self.sub_bg = sub_bg
        self.multichannel_ops = multichannel_ops

    @classmethod
    def default(cls):
        return cls({})

    @classmethod
    def from_meta(cls, meta):
        return cls(**exparams_from_meta(meta))


class Extractor(StepABC):
    """
    Apply a metric to cells identified in the tiles.

    Using the cell masks, the Extractor applies a metric, such as area or median, to cells identified in the image tiles.

    Its methods require both tile images and masks.

    Usually the metric is applied to only a tile's masked area, but some metrics depend on the whole tile.

    Extraction follows a three-level tree structure. Channels, such as GFP, are the root level; the reduction algorithm, such as maximum projection, is the second level; the specific metric, or operation, to apply to the masks, such as mean, is the third level.
    """

    # TODO Alan: Move this to a location with the SwainLab defaults
    default_meta = {
        "pixel_size": 0.236,
        "z_size": 0.6,
        "spacing": 0.6,
    }

    def __init__(
        self,
        parameters: ExtractorParameters,
        store: t.Optional[str] = None,
        tiler: t.Optional[Tiler] = None,
    ):
        """
        Initialise Extractor.

        Parameters
        ----------
        parameters: core.extractor Parameters
            Parameters that include the channels, reduction and
            extraction functions.
        store: str
            Path to the h5 file containing the cell masks.
        tiler: pipeline-core.core.segmentation tiler
            Class that contains or fetches the images used for segmentation.
        """
        self.params = parameters
        if store:
            self.local = store
            self.load_meta()
        else:
            # if no h5 file, use the parameters directly
            self.meta = {"channel": parameters.to_dict()["tree"].keys()}
        if tiler:
            self.tiler = tiler
        self.load_funs()

    @classmethod
    def from_tiler(
        cls,
        parameters: ExtractorParameters,
        store: str,
        tiler: Tiler,
    ):
        """Initiate from a tiler instance."""
        return cls(parameters, store=store, tiler=tiler)

    @classmethod
    def from_img(
        cls,
        parameters: ExtractorParameters,
        store: str,
        img_meta: tuple,
    ):
        """Initiate from images."""
        return cls(parameters, store=store, tiler=Tiler(*img_meta))

    @property
    def channels(self):
        """Get a tuple of the available channels."""
        if not hasattr(self, "_channels"):
            if type(self.params.tree) is dict:
                self._channels = tuple(self.params.tree.keys())
        return self._channels

    @property
    def current_position(self):
        return str(self.local).split("/")[-1][:-3]

    @property
    def group(self):
        """Return path within the h5 file."""
        if not hasattr(self, "_out_path"):
            self._group = "/extraction/"
        return self._group

    def load_custom_funs(self):
        """
        Incorporate the extra arguments of custom functions into their definitions.

        Normal functions only have cell_masks and trap_image as their
        arguments, and here custom functions are made the same by
        setting the values of their extra arguments.

        Any other parameters are taken from the experiment's metadata and automatically applied. These parameters therefore must be loaded within an Extractor instance.
        """
        # find functions specified in params.tree
        funs = set(
            [
                fun
                for ch in self.params.tree.values()
                for red in ch.values()
                for fun in red
            ]
        )
        # consider only those already loaded from CUSTOM_FUNS
        funs = funs.intersection(CUSTOM_FUNS.keys())
        # find their arguments
        self._custom_arg_vals = {
            k: {k2: self.get_meta(k2) for k2 in v}
            for k, v in CUSTOM_ARGS.items()
        }
        # define custom functions
        self._custom_funs = {}
        for k, f in CUSTOM_FUNS.items():

            def tmp(f):
                # pass extra arguments to custom function
                # return a function of cell_masks and trap_image
                return lambda cell_masks, trap_image: trap_apply(
                    f,
                    cell_masks,
                    trap_image,
                    **self._custom_arg_vals.get(k, {}),
                )

            self._custom_funs[k] = tmp(f)

    def load_funs(self):
        """Define all functions, including custum ones."""
        self.load_custom_funs()
        self._all_cell_funs = set(self._custom_funs.keys()).union(CELL_FUNS)
        # merge the two dicts
        self._all_funs = {**self._custom_funs, **FUNS}

    def load_meta(self):
        """Load metadata from h5 file."""
        self.meta = load_attributes(self.local)

    def get_tiles(
        self,
        tp: int,
        channels: t.Optional[t.List[t.Union[str, int]]] = None,
        z: t.Optional[t.List[str]] = None,
        **kwargs,
    ) -> t.Optional[np.ndarray]:
        """
        Find tiles for a given time point, channels, and z-stacks.

        Any additional keyword arguments are passed to tiler.get_tiles_timepoint

        Parameters
        ----------
        tp: int
            Time point of interest.
        channels: list of strings (optional)
            Channels of interest.
        z: list of integers (optional)
            Indices for the z-stacks of interest.
        """
        if channels is None:
            # find channels from tiler
            channel_ids = list(range(len(self.tiler.channels)))
        elif len(channels):
            # a subset of channels was specified
            channel_ids = [self.tiler.get_channel_index(ch) for ch in channels]
        else:
            # a list of the indices of the z stacks
            channel_ids = None
        if z is None:
            # gets the tiles data via tiler
            z = list(range(self.tiler.shape[-3]))
        res = (
            self.tiler.get_tiles_timepoint(
                tp, channels=channel_ids, z=z, **kwargs
            )
            if channel_ids
            else None
        )
        # data arranged as (tiles, channels, time points, X, Y, Z)
        return res

    def extract_traps(
        self,
        traps: t.List[np.ndarray],
        masks: t.List[np.ndarray],
        metric: str,
        labels: t.Dict[int, t.List[int]],
    ) -> t.Tuple[t.Union[t.Tuple[float], t.Tuple[t.Tuple[int]]]]:
        """
        Apply a function to a whole position.

        Parameters
        ----------
        traps: list of arrays
            t.List of images.
        masks: list of arrays
            t.List of masks.
        metric: str
            Metric to extract.
        labels: dict
            A dict of cell labels with trap_ids as keys and a list of cell labels as values.
        pos_info: bool
            Whether to add the position as an index or not.

        Returns
        -------
        res_idx: a tuple of tuples
            A two-tuple comprising a tuple of results and a tuple of the tile_id and cell labels
        """
        if labels is None:
            self._log("No labels given. Sorting cells using index.")
        cell_fun = True if metric in self._all_cell_funs else False
        idx = []
        results = []
        for trap_id, (mask_set, trap, lbl_set) in enumerate(
            zip(masks, traps, labels.values())
        ):
            # ignore empty traps
            if len(mask_set):
                # apply metric either a cell function or otherwise
                result = self._all_funs[metric](mask_set, trap)
                if cell_fun:
                    # store results for each cell separately
                    for lbl, val in zip(lbl_set, result):
                        results.append(val)
                        idx.append((trap_id, lbl))
                else:
                    # background (trap) function
                    results.append(result)
                    idx.append(trap_id)
        res_idx = (tuple(results), tuple(idx))
        return res_idx

    def extract_funs(
        self,
        traps: t.List[np.array],
        masks: t.List[np.array],
        metrics: t.List[str],
        **kwargs,
    ) -> t.Dict[str, pd.Series]:
        """
        Return dict with metrics as key and metrics applied to data as values.

        Data from one time point is used.
        """
        d = {
            metric: self.extract_traps(
                traps=traps, masks=masks, metric=metric, **kwargs
            )
            for metric in metrics
        }
        return d

    def reduce_extract(
        self,
        traps: np.ndarray,
        masks: t.List[np.ndarray],
        red_metrics: t.Dict[reduction_method, t.Collection[str]],
        **kwargs,
    ) -> t.Dict[str, t.Dict[reduction_method, t.Dict[str, pd.Series]]]:
        """
        Wrapper to apply reduction and then extraction.

        Parameters
        ----------
        tiles_data: array
            An array of image data arranged as (tiles, X, Y, Z)
        masks: list of arrays
            An array of masks for each trap: one per cell at the trap
        red_metrics: dict
            dict for which keys are reduction functions and values are either a list or a set of strings giving the metric functions.
            For example: {'np_max': {'max5px', 'mean', 'median'}}
        **kwargs: dict
            All other arguments passed to Extractor.extract_funs.

        Returns
        ------
        Dict of dataframes with the corresponding reductions and metrics nested.
        """
        # create dict with keys naming the reduction in the z-direction and the reduced data as values
        reduced_tiles_data = {}
        if traps is not None:
            for red_fun in red_metrics.keys():
                reduced_tiles_data[red_fun] = [
                    self.reduce_dims(tile_data, method=RED_FUNS[red_fun])
                    for tile_data in traps
                ]
        d = {
            red_fun: self.extract_funs(
                metrics=metrics,
                traps=reduced_tiles_data.get(red_fun, [None for _ in masks]),
                masks=masks,
                **kwargs,
            )
            for red_fun, metrics in red_metrics.items()
        }
        return d

    def reduce_dims(
        self, img: np.ndarray, method: reduction_method = None
    ) -> np.ndarray:
        """
        Collapse a z-stack into 2d array using method.

        If method is None, return the original data.

        Parameters
        ----------
        img: array
            An array of the image data arranged as (X, Y, Z).
        method: function
            The reduction function.
        """
        reduced = img
        if method is not None:
            reduced = reduce_z(img, method)
        return reduced

    def extract_tp(
        self,
        tp: int,
        tree: t.Optional[extraction_tree] = None,
        tile_size: int = 117,
        masks: t.Optional[t.List[np.ndarray]] = None,
        labels: t.Optional[t.List[int]] = None,
        **kwargs,
    ) -> t.Dict[str, t.Dict[str, t.Dict[str, tuple]]]:
        """
        Extract for an individual time point.

        Parameters
        ----------
        tp : int
            Time point being analysed.
        tree : dict
            Nested dictionary indicating channels, reduction functions and
            metrics to be used.
            For example: {'general': {'None': ['area', 'volume', 'eccentricity']}}
        tile_size : int
            Size of the tile to be extracted.
        masks : list of arrays
            A list of masks per trap with each mask having dimensions (ncells, tile_size,
            tile_size).
        labels : dict
            A dictionary with trap_ids as keys and cell_labels as values.
        **kwargs : keyword arguments
            Passed to extractor.reduce_extract.

        Returns
        -------
        d: dict
            Dictionary of the results with three levels of dictionaries.
            The first level has channels as keys.
            The second level has reduction metrics as keys.
            The third level has cell or background metrics as keys and a two-tuple as values.
            The first tuple is the result of applying the metrics to a particular cell or trap; the second tuple is either (trap_id, cell_label) for a metric applied to a cell or a trap_id for a metric applied to a trap.

            An example is d["GFP"]["np_max"]["mean"][0], which gives a tuple of the calculated mean GFP fluorescence for all cells.
        """
        # TODO Can we split the different extraction types into sub-methods to make this easier to read?
        if tree is None:
            # use default
            tree: extraction_tree = self.params.tree
        # dictionary with channel: {reduction algorithm : metric}
        ch_tree = {ch: v for ch, v in tree.items() if ch != "general"}
        # tuple of the channels
        tree_chs = (*ch_tree,)
        # create a Cells object to extract information from the h5 file
        cells = Cells(self.local)
        # find the cell labels and store as dict with trap_ids as keys
        if labels is None:
            raw_labels = cells.labels_at_time(tp)
            labels = {
                trap_id: raw_labels.get(trap_id, [])
                for trap_id in range(cells.ntraps)
            }
        # find the cell masks for a given trap as a dict with trap_ids as keys
        if masks is None:
            raw_masks = cells.at_time(tp, kind="mask")
            masks = {trap_id: [] for trap_id in range(cells.ntraps)}
            for trap_id, cells in raw_masks.items():
                if len(cells):
                    masks[trap_id] = np.stack(np.array(cells)).astype(bool)
        # convert to a list of masks
        masks = [np.array(v) for v in masks.values()]
        # find image data at the time point
        # stored as an array arranged as (traps, channels, time points, X, Y, Z)
        tiles = self.get_tiles(tp, tile_shape=tile_size, channels=tree_chs)
        # generate boolean masks for background as a list with one mask per trap
        bgs = np.array([])
        if self.params.sub_bg:
            bgs = ~np.array(
                list(
                    map(
                        lambda x: np.sum(x, axis=0)
                        if np.any(x)
                        else np.zeros((tile_size, tile_size)),
                        masks,
                    )
                )
            ).astype(bool)
        # perform extraction by applying metrics
        d = {}
        self.img_bgsub = {}
        for ch, red_metrics in tree.items():
            # NB ch != is necessary for threading
            if ch != "general" and tiles is not None and len(tiles):
                # image data for all traps and z sections for a particular channel
                # as an array arranged as (tiles, Z, X, Y, )
                img = tiles[:, tree_chs.index(ch), 0]
            else:
                img = None
            # apply metrics to image data
            d[ch] = self.reduce_extract(
                traps=img,
                masks=masks,
                red_metrics=red_metrics,
                labels=labels,
                **kwargs,
            )
            # apply metrics to image data with the background subtracted
            if bgs.any() and ch in self.params.sub_bg and img is not None:
                # calculate metrics with subtracted bg
                ch_bs = ch + "_bgsub"
                # subtract median background

                self.img_bgsub[ch_bs] = np.moveaxis(
                    np.stack(
                        list(
                            map(
                                lambda tile, mask: np.moveaxis(tile, 0, -1)
                                - bn.median(tile[:, mask], axis=1),
                                img,
                                bgs,
                            )
                        )
                    ),
                    -1,
                    1,
                )  # End with tiles, z, y, x
                # apply metrics to background-corrected data
                d[ch_bs] = self.reduce_extract(
                    red_metrics=ch_tree[ch],
                    traps=self.img_bgsub[ch_bs],
                    masks=masks,
                    labels=labels,
                    **kwargs,
                )
        # apply any metrics using multiple channels, such as pH calculations
        for name, (
            chs,
            merge_fun,
            red_metrics,
        ) in self.params.multichannel_ops.items():
            if len(
                set(chs).intersection(
                    set(self.img_bgsub.keys()).union(tree_chs)
                )
            ) == len(chs):
                channels_stack = np.stack(
                    [self.get_imgs(ch, tiles, tree_chs) for ch in chs], axis=-1
                )
                merged = RED_FUNS[merge_fun](channels_stack, axis=-1)
                d[name] = self.reduce_extract(
                    red_metrics=red_metrics,
                    traps=merged,
                    masks=masks,
                    labels=labels,
                    **kwargs,
                )
        return d

    def get_imgs(self, channel: t.Optional[str], tiles, channels=None):
        """
        Return image from a correct source, either raw or bgsub.

        Parameters
        ----------
        channel: str
            Name of channel to get.
        tiles: ndarray
            An array of the image data having dimensions of (tile_id, channel, tp, tile_size, tile_size, n_zstacks).
        channels: list of str (optional)
            t.List of available channels.

        Returns
        -------
        img: ndarray
            An array of image data with dimensions (no tiles, X, Y, no Z channels)
        """
        if channels is None:
            channels = (*self.params.tree,)
        if channel in channels:  # TODO start here to fetch channel using regex
            return tiles[:, channels.index(channel), 0]
        elif channel in self.img_bgsub:
            return self.img_bgsub[channel]

    def _run_tp(
        self,
        tps: t.List[int] = None,
        tree=None,
        save=True,
        **kwargs,
    ) -> dict:
        """
        Wrapper to add compatibility with other steps of the pipeline.

        Parameters
        ----------
        tps: list of int (optional)
            Time points to include.
        tree: dict (optional)
            Nested dictionary indicating channels, reduction functions and
            metrics to be used.
            For example: {'general': {'None': ['area', 'volume', 'eccentricity']}}
        save: boolean (optional)
            If True, save results to h5 file.
        kwargs: keyword arguments (optional)
            Passed to extract_tp.

        Returns
        -------
        d: dict
            A dict of the extracted data with a concatenated string of channel, reduction metric, and cell metric as keys and pd.Series of the extracted data as values.
        """
        if tree is None:
            tree = self.params.tree
        if tps is None:
            tps = list(range(self.meta["time_settings/ntimepoints"][0]))
        elif isinstance(tps, int):
            tps = [tps]
        # store results in dict
        d = {}
        for tp in tps:
            # extract for each time point and convert to dict of pd.Series
            new = flatten_nesteddict(
                self.extract_tp(tp=tp, tree=tree, **kwargs),
                to="series",
                tp=tp,
            )
            # concatenate with data extracted from early time points
            for k in new.keys():
                d[k] = pd.concat((d.get(k, None), new[k]), axis=1)
        # add indices to pd.Series containing the extracted data
        for k in d.keys():
            indices = ["experiment", "position", "trap", "cell_label"]
            idx = (
                indices[-d[k].index.nlevels :]
                if d[k].index.nlevels > 1
                else [indices[-2]]
            )
            d[k].index.names = idx
        # save
        if save:
            self.save_to_hdf(d)
        return d

    def save_to_hdf(self, dict_series, path=None):
        """
        Save the extracted data to the h5 file.

        Parameters
        ----------
        dict_series: dict
            A dictionary of the extracted data, created by run.
        path: Path (optional)
            To the h5 file.
        """
        if path is None:
            path = self.local
        self.writer = Writer(path)
        for extract_name, series in dict_series.items():
            dset_path = "/extraction/" + extract_name
            self.writer.write(dset_path, series)
        self.writer.id_cache.clear()

    def get_meta(self, flds: t.Union[str, t.Collection]):
        """Obtain metadata for one or multiple fields."""
        if isinstance(flds, str):
            flds = [flds]
        meta_short = {k.split("/")[-1]: v for k, v in self.meta.items()}
        return {
            f: meta_short.get(f, self.default_meta.get(f, None)) for f in flds
        }


### Helpers
def flatten_nesteddict(
    nest: dict, to="series", tp: int = None
) -> t.Dict[str, pd.Series]:
    """
    Convert a nested extraction dict into a dict of pd.Series.

    Parameters
    ----------
    nest: dict of dicts
        Contains the nested results of extraction.
    to: str (optional)
        Specifies the format of the output, either pd.Series (default) or a list
    tp: int
        Time point used to name the pd.Series

    Returns
    -------
    d: dict
        A dict with a concatenated string of channel, reduction metric, and cell metric as keys and either a pd.Series or a list of the corresponding extracted data as values.
    """
    d = {}
    for k0, v0 in nest.items():
        for k1, v1 in v0.items():
            for k2, v2 in v1.items():
                d["/".join((k0, k1, k2))] = (
                    pd.Series(*v2, name=tp) if to == "series" else v2
                )
    return d


class hollowExtractor(Extractor):
    """
    Extractor that only cares about receiving images and masks.

    Used for testing.
    """

    def __init__(self, parameters):
        self.params = parameters



## Trap-wise calculations

import numpy as np


def imBackground(cell_masks, trap_image):
    """
    Find the median background (pixels not comprising cells) from trap_image.

    Parameters
    ----------
    cell_masks: 3d array
       Segmentation masks for cells
    trap_image:
        The image (all channels) for the tile containing the cell.
    """
    if not len(cell_masks):
        # create cell_masks if none are given
        cell_masks = np.zeros_like(trap_image)
    # find background pixels
    # sum over all cells identified at a trap - one mask for each cell
    background = ~cell_masks.sum(axis=2).astype(bool)
    return np.median(trap_image[np.where(background)])


def background_max5(cell_masks, trap_image):
    """
    Finds the mean of the maximum five pixels of the background.

    Parameters
    ----------
    cell_masks: 3d array
        Segmentation masks for cells.
    trap_image:
        The image (all channels) for the tile containing the cell.
    """
    if not len(cell_masks):
        # create cell_masks if none are given
        cell_masks = np.zeros_like(trap_image)
    # find background pixels
    # sum over all cells identified at a trap - one mask for each cell
    background = ~cell_masks.sum(axis=2).astype(bool)
    return np.mean(np.sort(trap_image[np.where(background)])[-5:])
